# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from pathlib import Path
import pandas as pd
import numpy as np
from functools import reduce

import phenonaut.data
from typing import Optional, Tuple, Union
from phenonaut.data import dataset

from phenonaut.data.dataset import Dataset
from phenonaut.phenonaut import Phenonaut
from phenonaut.predict.predictor_dataclasses import PhenonautPredictionMetric, PhenonautPredictor
from enum import Enum, auto

# The PredictionType enum is used in multiple places to denote the prediction task.
# 'view' is simply multi-y-regression for prediction either different views, or multiple
# properties at the same time
class PredictionType(Enum):
    """PredictionType Enum for classification, regression or view prediction.

    Parameters
    ----------
    Enum : int
        Enumerated type, captures if the prediction task is classification,
        regression, or view (multiregression)
    """    
    classification = auto()
    regression = auto()
    view = auto()


def get_y_from_target(
    data: Union[phenonaut.data.Dataset, pd.DataFrame, list[phenonaut.data.Dataset], list[pd.DataFrame]],
    target: Optional[Union[str, pd.Series, np.ndarray, tuple]] = None,
) -> Union[list, pd.Series, tuple, np.ndarray]:
    """Get target y from Dataset(s) or DataFrame

    Parameters
    ----------
    data : Union[list[Dataset],list[DataFrame]]
        Dataset, DataFrame, list of Datasets or list of DataFrames where the
        target values are present.
    target : Optional[Union[str, Series, np.ndarray, tuple]], optional
        Target column name containing target y values in Dataset/DataFrame. If
        None, then the target y is attempted to be inferred by looking for
        Dataset columns which are not listed as features. Target y cannot be
        inferred from DataFrames. If np.ndarray, tuple, or pd.Series, then this
        is directly returned as the target of prediciton y, by default None.

    Returns
    -------
    Union[list, Series, tuple, np.ndarray]
        Target y values for prediction

    Raises
    ------
    ValueError
        Given target string (column title) was a feature of the dataset.
    ValueError
        Object should be phenonaut.Dataset or pd.DataFrame.
    ValueError
        Given target string (column title) was not found in any supplied
        Datasets/Dataframes.
    ValueError
        Target was not set, trying to guess it from a phenonaut.Dataset, but
        did not find this type.
    ValueError
        Could not guess the target from supplied Dataset(s). Pass target as
        string for a dataset.df column heading, or the prediction target
        directly.
    """
    # If target is directly given (series, ndarray, tuple, list), then return it

    if isinstance(target, (pd.Series, np.ndarray, tuple, list)):
        return target
    if isinstance(target, Dataset):
        return target.data

    # Processing bellow expects data to be in a list Data should always be a list
    if not isinstance(data, list):
        data = [data]

    # If target is given, use it to set y.
    if target is not None:
        # If string, then search through the datasets to find a column with that value, use the first found
        if isinstance(target, str):
            for d in data:
                if isinstance(d, pd.DataFrame):
                    if target in d.columns.values:
                        return d[target]
                elif isinstance(d, phenonaut.data.Dataset):
                    if target in d.features:
                        raise (
                            f"Given target string (column title) was {target}, but this is a feature of the dataset (name={d.name}, features={d.features}"
                        )
                    if target in d.get_non_feature_columns():
                        return d.df[target]
                else:
                    raise ValueError(f"Object of type {type(d)} found, should be phenonaut.Dataset or pd.DataFrame")
            raise ValueError(
                f"Given target string (column title) was {target}, but this was not found in any supplied Datasets/Dataframes"
            )

    # Target was not set, guess it by identifying a ds that has one column not listed as a feature.
    for d in data:
        if not isinstance(d, phenonaut.data.Dataset):
            raise ValueError(f"Target was not set, trying to guess it from a phenonaut.Dataset, but found {type(d)}")
        if "prediction_target" in d._metadata.keys():
            return d.df[d._metadata["prediction_target"]]
        non_feature_columns = d.get_non_feature_columns()
        if len(non_feature_columns) == 1:
            return d.df[non_feature_columns[0]]
    raise ValueError(
        "Could not guess the target from supplied Dataset(s). Pass target as string for a dataset.df column heading, or the prediction target directly"
    )


def get_prediction_type_from_y(y):
    """For a given target y - get prediction type

    Looking at the data in y, return prediction type from classification,
    regression or multiregression (view prediction).

    Parameters
    ----------
    y : _type_
        _description_

    Returns
    -------
    _type_
        _description_

    Raises
    ------
    ValueError
        _description_
    """    
    if isinstance(y, pd.Series):
        if y.values.dtype in [int, np.int64]:
            return PredictionType.classification
        else:
            return PredictionType.regression
    if isinstance(y, list):
        if type(y[0]) == int:
            return PredictionType.classification
        else:
            return PredictionType.regression
    if isinstance(y, np.ndarray):
        if y.ndim == 1:
            if y.dtype == int:
                return PredictionType.classification
            else:
                return PredictionType.regression
        else:
            return PredictionType.view
    if isinstance(y, pd.DataFrame):
        if y.values.shape[1] > 1:
            return PredictionType.view
        else:
            if y.iloc[0, 0].values.dtype == int:
                return PredictionType.classification
            else:
                return PredictionType.regression
    if isinstance(y, phenonaut.data.Dataset):
        return PredictionType.view
    raise ValueError(f"y was not pd.Series, list, np.array, phenonaut.data.Dataset, it was : {type(y)}")


def get_common_indexes(dataframes_list: list[pd.DataFrame]) -> list[str]:
    """Get common indexes from list of DataFrames

    Parameters
    ----------
    dataframes_list : list[pd.DataFrame]
        List of pd.DataFrames from which common indexes should be extracted.

    Returns
    -------
    list[str]
        List of common indexes between pd.DataFrames.
    """    
    return list(reduce(lambda x, y: x & y, [set(df.index) for df in dataframes_list]))


def get_df_from_optuna_db(
    optuna_db_file: Union[str, Path],
    csv_output_filename: Union[Path, str] = None,
    json_output_filename: Union[Path, str] = None,
    get_only_best_per_study: bool = False,
) -> pd.DataFrame:
    """After predict.profile, turn Optuna sqlite3 files into pd.DataFrame

    Parameters
    ----------
    optuna_db_file : Union[str, Path]
        Optuna hyperparameter optimisation database (sqlite3file)
    csv_output_filename : Union[Path, str], optional
        Target output CSV file. Can be None, in which case, no CSV file is
        written out. By default None.
    json_output_filename : Union[Path, str], optional
        Target output JSON file. Can be None, in which case, no JSON file is
        written out. By default None.
    get_only_best_per_study : bool, optional
        Boolean value stating if only the best hyperparameter set per study
        should be writen out. By default False.

    Returns
    -------
    pd.DataFrame
        DataFrame sumarising Optuna hyperparameter scan results.

    Raises
    ------
    FileNotFoundError
        Database file (sqlite3) not found.
    """
    import optuna

    df_dict_list = []
    if isinstance(optuna_db_file, str):
        optuna_db_file = Path(optuna_db_file)
    if not optuna_db_file.exists():
        raise FileNotFoundError(f"Could not find optuna database file ({optuna_db_file})")
    study_names = [
        study.study_name
        for study in optuna.study.get_all_study_summaries(storage=f"sqlite:///{optuna_db_file.resolve()}")
    ]
    for study_name in study_names:
        study = optuna.study.load_study(study_name, storage=f"sqlite:///{optuna_db_file.resolve()}")
        if get_only_best_per_study:
            # If optuna is interuped with CTRL+C when using an sqlite database, then an sqlite database
            # can have a study, but no trials, resulting in a crash here. The following issue addresses
            # the problem:  https://github.com/optuna/optuna/issues/2101
            try:
                trial = study.best_trial
            except ValueError:
                # Optuna/sqlite returned
                # ValueError: Record does not exist.
                # Study does not contain any trials
                continue
            row_dict = {k: v for k, v in study.user_attrs.items()}
            row_dict.update({k: v for k, v in trial.user_attrs.items()})

            if "KFoldScores" in row_dict:
                del row_dict["KFoldScores"]
                fold_scores = trial.user_attrs["KFoldScores"]
                row_dict.update({f"FoldScore_{i+1}": fold_scores[i] for i in range(len(fold_scores))})
            row_dict.update({"parameters": trial.params})
            df_dict_list.append(row_dict)

        else:  # Getting all
            for trial in study.get_trials():
                row_dict = {
                    k: v
                    for k, v in study.user_attrs.items()
                    if k not in ["best_trained_model_score", "best_trained_model"]
                }
                row_dict.update({k: v for k, v in trial.user_attrs.items()})
                if "KFoldScores" in row_dict:
                    del row_dict["KFoldScores"]
                    fold_scores = trial.user_attrs["KFoldScores"]
                    row_dict.update({f"FoldScore_{i+1}": fold_scores[i] for i in range(len(fold_scores))})
                row_dict.update({"parameters": trial.params})
                df_dict_list.append(row_dict)

    df = pd.DataFrame(df_dict_list)
    columns_for_the_end = ["test_score", "parameters"]
    df = df[
        [colname for colname in df if not colname in columns_for_the_end]
        + [endcol for endcol in columns_for_the_end if endcol in df]
    ]
    if csv_output_filename is not None:
        csv_output_filename = Path(csv_output_filename)
        if not csv_output_filename.parent.exists():
            csv_output_filename.mkdir(parents=True)
        df.to_csv(csv_output_filename)
    if json_output_filename is not None:
        json_output_filename = Path(json_output_filename)
        if not json_output_filename.parent.exists():
            json_output_filename.mkdir(parents=True)
        df.to_json(json_output_filename)
    return df


def get_best_predictor_dataset_df(df: pd.DataFrame, column_containing_values:str="test_score") -> pd.DataFrame:
    """For a given Optuna hyperaprameter scan dataframe, get the best predictor

    Parameters
    ----------
    df : pd.DataFrame
        Optua hyperparameterscan pd.DataFrame, likely generated by
        get_df_from_optuna_db.
    column_containing_values : str, optional
        Name of the column containing scores. By default "test_score".

    Returns
    -------
    pd.DataFrame
        DataFrame containing information on the best predictor.
    """    
    unique_predictors = df["predictor_name"].unique()
    unique_dataset_views = df["dataset"].unique()
    heatmap_data_array = np.full((len(unique_dataset_views), len(unique_predictors)), np.nan)

    for pred_i in range(len(unique_predictors)):
        for ds_i in range(len(unique_dataset_views)):
            heatmap_data_array[ds_i, pred_i] = df.query(
                f"predictor_name=='{unique_predictors[pred_i]}' and dataset=='{str(unique_dataset_views[ds_i])}'"
            )[column_containing_values].max()

    return pd.DataFrame(heatmap_data_array, index=unique_dataset_views, columns=unique_predictors)




def get_metric(metric: Union[str, dict, PhenonautPredictionMetric]):
    """Get metric function from various options for metric definition

    Helper function which allows specification of metrics with strings
    indicating common names, dictionaries.

    Currently understands the shortcut strings:
    accuracy, accuracy_score
    mse, MSE, mean_squared_error
    rmse, RMSE, root_mean_squared_error
    AUROC, auroc, area_under_roc_curve

    Parameters
    ----------
    metric : Union[str, dict, PhenonautPredictionMetric]
        String, dict or PhenonautPredictionMetric to be used for scoring

    Returns
    -------
    PhenonautPredictionMetric
        Prediction metric

    Raises
    ------
    ValueError
        No metrics found matching short string name.
    KeyError
        Given dictionary did not include all required fields.
    ValueError
        metric argument was not of a suitable type.
    """    
    if isinstance(metric, PhenonautPredictionMetric):
        return metric
    elif isinstance(metric, str):
        if metric in ["accuracy", "accuracy_score"]:
            from sklearn.metrics import accuracy_score

            return PhenonautPredictionMetric(accuracy_score, "Accuracy", False)

        elif metric in ["mse", "MSE", "mean_squared_error"]:
            from sklearn.metrics import mean_squared_error

            return PhenonautPredictionMetric(mean_squared_error, "MSE", True)

        elif metric in ["rmse", "RMSE", "root_mean_squared_error"]:
            from sklearn.metrics import mean_squared_error

            return PhenonautPredictionMetric(lambda x, y: mean_squared_error(x, y, squared=False), "RMSE", True)

        elif metric in ["AUROC", "auroc", "area_under_roc_curve"]:
            from sklearn.metrics import roc_auc_score

            return PhenonautPredictionMetric(roc_auc_score, "AUROC", False)

        raise ValueError(
            f"Did not find any predefined metrics matching the given '{metric}' shortcut, try accuracy, mse, rmse, or auroc"
        )

    elif isinstance(metric, dict):
        if not all([k in metric for k in ("func", "name", "lower_is_better")]):
            raise KeyError(
                f"'func', 'name', and 'lower_is_better' should be metric dictionary keys (containing the types Callable, str, bool), but found: {', '.join(metric.keys())}"
            )
        return PhenonautPredictionMetric(metric["func"], metric["name"], metric["lower_is_better"])
    else:
        raise ValueError(f"Metric should be a str, PhenonautPredictionMetric, or dict, it was: {type(metric)}")


def get_X_y(
    phe: Phenonaut,
    dataset_combination: Union[tuple, list],
    target,
    predictor: PhenonautPredictor,
    prediction_type: PredictionType,
) -> tuple:
    """For a given set of views, and known y, get X and y for predictor training

    Parameters
    ----------
    phe : Phenonaut
        The Phenonaut object containing Datasets
    dataset_combination : Union[tuple, list]
        Dataset combinations to be used in prediction task.
    target : pd.Series, np.ndarray
        Prediction target
    predictor : PhenonautPredictor
        The PhenonautPredictor being used.
    prediction_type : PredictionType
        Enum classification type specifying classification, regression or view.

    Returns
    -------
    tuple
        X, y tuple for training of predictor.
    """
    X = None
    y = get_y_from_target(phe[dataset_combination], target=target)
    common_indices = None
    if isinstance(y, (pd.DataFrame, pd.Series)):
        common_indices = get_common_indexes([phe[dsn].df for dsn in dataset_combination] + [y])
    elif isinstance(y, dataset):
        common_indices = get_common_indexes([phe[dsn].df for dsn in dataset_combination] + [y.data])
    # 1 datasets
    if len(dataset_combination) == 1:
        #  1 view predictor
        if predictor.num_views == 1:
            if prediction_type == PredictionType.view:
                if common_indices is None:
                    return (phe[dataset_combination[0]].data.values, y.values)
                else:
                    return (
                        phe[dataset_combination[0]].data.loc[common_indices, :].values,
                        y.loc[common_indices, :].values,
                    )
            else:
                if common_indices is None:
                    return (phe[dataset_combination[0]].data.values, y.values)
                else:
                    return (
                        phe[dataset_combination[0]].data.loc[common_indices, :].values,
                        y.loc[common_indices].values,
                    )

        # Multiview predictor not compatible with single view data
        else:
            return (None, None)

    # Multiple datasets - multiview
    else:
        if predictor.num_views == 1:  # Predictor can only do 1 view
            if common_indices is None:
                return (np.hstack([phe[ds_name].data.values for ds_name in dataset_combination]), y.values)
            else:
                return (
                    np.hstack([phe[ds_name].data.loc[common_indices, :].values for ds_name in dataset_combination]),
                    y.loc[common_indices].values,
                )
        else:  # Predictor is multiview
            if len(dataset_combination) != predictor.num_views:
                return (None, None)
            if common_indices is None:
                return ([phe[ds_name].data.values for ds_name in dataset_combination], y.values)
            else:
                return (
                    [phe[ds_name].data.loc[common_indices, :].values for ds_name in dataset_combination],
                    y.loc[common_indices].values,
                )



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from .phenonaut import Phenonaut
import phenonaut.metrics
import phenonaut.output
from phenonaut.data import PlatemapQuerier
import phenonaut.packaged_datasets



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from .tcga import TCGA
from .cmap import CMAP
from .iris import Iris
from .iris import Iris_2_views
from .breast_cancer import BreastCancer





# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

kinds = {
    "drugseq": {
        "transforms": [
            ("replace_str", ("Var2", "Pos_", "Pos-")),
            ("replace_str", ("Var2", "Neg_", "Neg-")),
            ("split_column", ("Var2", "_", ["WellName", "CpdID", "BARCODE"])),
            ("replace_str", "Var2", "Pos-", "Pos_"),
            ("replace_str", "Var2", "Neg-", "Neg_"),
            ("pivot", "Var1", "value"),
        ],
        "features_prefix": "ENSG",
        "index_col": 0,
    },
    
    "feat":{
        'features_prefix':'feat_'
    }
}




# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from abc import ABC
from typing import Optional, Union
import pandas as pd
from sklearn.preprocessing import StandardScaler as sklearn_StandardScaler
from phenonaut.data import Dataset
import numpy as np


class GenericTransform(ABC):
    def __init__(self):
        pass

    def __call__(
        self,
    ):
        pass


class StandardScaler(GenericTransform):
    def __init__(self):
        """StandardScaler
        
        Perform StandardScaler on a dataset
        
        May be instantiated and then called, or fit and then transform performed.
        Optional query argument carries out a pandas query to select relevant rows
        from the dataframe to perform the fit.

        """
        self._sc=sklearn_StandardScaler()
    def fit(self, dataset:Dataset, query:Optional[str]=None):
        """Fit the standard scaler to a dataset

        Parameters
        ----------
        dataset : Dataset
            Dataset upon which the fit is to be performed
        query : Optional[str], optional
            Pandas query which may be used to select rows, by default None

        Returns
        -------
        StandardScaler
            Returns a reference to itself as an object.
        """        
        if query is None:
            self._sc.fit(dataset.data)
        else:
            self._sc.fit(dataset.df.query(query)[dataset.features])
        return self
    def transform(self, dataset:Dataset):
        """Perform transform

        Parameters
        ----------
        dataset : Dataset
            The dataset upon which the transformation should be applied.
        """        
        dataset.df[dataset.features] = self._sc.fit_transform(
            dataset.data
        )
        dataset.features=dataset.features, "Applied standard scalar"
    def __call__(self, dataset: Dataset, query:Optional[str]=None):
        """Perform StandardScaler

        Object may be called directly to perform fit and then transform.
        Parameters
        ----------
        dataset : Dataset
            The dataset upon which the transformation should be applied.
        query : Optional[str], optional
            Pandas query which may be used to select rows, by default None
        """        
        self.fit(dataset, query)
        self.transform(dataset)


class Log2(GenericTransform):
    def __call__(self, dataset: Dataset):
        super().__init__()
        dataset.df[dataset.features] = np.log2(dataset.data)

class RobustMAD(GenericTransform):
    def __init__(self, mad_scale:Union[str, float]='normal', epsilon:Optional[float]=None):
        """RobustMAD
        
        Robust scaling, as used popularly in pycytominer

        https://github.com/cytomining/pycytominer/blob/master/pycytominer/operations/transform.py

        Perform robust normalization using median and mad
        (median absolute deviation), like:
        
        X=X-median/(mad+epsilon)
        
        May be instantiated and then called, or fit and then transform performed.
        Optional query argument carries out a pandas query to select relevant rows
        from the dataframe to perform the fit.

        Parameters
        ----------
        mad_scale : Union[str,float], optional
            Scaling argument passed to the median_abs_deviation 
            scipy function. The value 'normal' implies the data
            is normally distributed, but this may be changed
            for other distribution types. See scipy documentation
            for the function for a more complete explanation.
            By default 'normal'.
        epsilon : float, optional
            Small fudge-factor to remove danger of median value.
            If None, then it is set by numpy eps value.
            By default None.
        """        
        self._epsilon=np.finfo(np.float64).eps if epsilon is None else epsilon
        self._mad_scale=mad_scale

    def fit(self, dataset:Dataset, query:Optional[str]=None):
        """Fit the RobustMAD scaler to a dataset

        Parameters
        ----------
        dataset : Dataset
            Dataset upon which the fit is to be performed
        query : Optional[str], optional
            Pandas query which may be used to select rows, by default None

        Returns
        -------
        RobustMAD
            Returns a reference to itself as an object.
        """        
        from scipy.stats import median_abs_deviation
        if query is None:
            self._median=dataset.data.median()
            self._mad=pd.Series(median_abs_deviation(dataset.data, nan_policy="omit", scale=self._mad_scale), index=self._median.index)
        else:
            query_df=dataset.df.query(query)[dataset.features]
            self._median=query_df.median()
            self._mad=pd.Series(median_abs_deviation(query_df, nan_policy="omit", scale=self._mad_scale), index=self._median.index)
        return self
    def transform(self, dataset:Dataset):
        """Perform transform

        Parameters
        ----------
        dataset : Dataset
            The dataset upon which the transformation should be applied.
        """        
        dataset.df[dataset.features]=(dataset.data-self._median)/(self._mad+self._epsilon)
    def __call__(self, dataset:Dataset, query:Optional[str]=None):
        """Perform RobustMAD

        Object may be called directly to perform fit and then transform.
        Parameters
        ----------
        dataset : Dataset
            The dataset upon which the transformation should be applied.
        query : Optional[str], optional
            Pandas query which may be used to select rows, by default None
        """        
        self.fit(dataset, query)
        self.transform(dataset)




# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from scipy.spatial.distance import cdist
from inspect import signature
from pathlib import Path
from typing import Callable, Optional, Union
import pandas as pd
import numpy as np
from pandas.errors import DataError

from copy import deepcopy
from itertools import product as itertools_product
from collections import namedtuple
from phenonaut import data
from phenonaut.utils import load_dict
from hashlib import sha256
import re

TransformationHistory = namedtuple("TransformationHistory", ["features", "description"])


class Dataset:

    """Dataset constructor

    Dataset holds source agnostic datasets, read in using
    hints from a user configurable YAML file describing the input CSV file
    format and indicating key columns.

    Parameters
    ----------
    dataset_name : str
        Dataset name
    input_file_path : Union[Path, str, pd.DataFrame]
        Location of the input CSV/TSV/H5 file to be read, or a pd.DataFrame to use.
        If None, then an empty DataFrame object is returned. As well as CSV/TSV files,
        the location of an h5 file may also be given which contains a pandas dataframe.
        If h5 is given, then it is expected that a h5_key argument be passed.
    metadata : Union[dict, Path, str]
        Dictionary or path to yml file describing CSV file format and key
        columns. a 'sep' key:value pair may be supplied, but if absent, then the file
        is examined and if a TAB character is found present in the first line of the file,
        then it is assumed that the TAB character should be used to delimit values. This check
        is not performed if a 'sep' key is found in metadata, allowing a simple way to
        override this check.
    kind : Optional[str]
        Instead of providing metadata, some presets are available, which make reading
        in things like DRUG-Seq easier, without the need to explicitly set all
        required transforms. If used as well as metadata, then the preset metadata
        dictionary from the kind argument is first loaded, then updated with anything in
        the metadata dictionary, this therefore allows overriding specific presets present
        in kind dictionaries. Available 'kind' dictionaries may be listed by examining:
        phenonaut.data.recipes.recipes.keys()
    init_hash : Optional[Union[str, bytes]]
        Cryptographic hashing within Phenonaut Datasets can be initialised
        with a starting/seed hash. This is useful in the creation of
        blockchain-like chains of hashes. In environments where timestamping
        is unavailable, hashes may be published and then used as input to
        subsequent experiments. Building up a provable chain along the way.
        By default None, implying an empty bytes array.
    h5_key : Optional[str]
            If input_file_path is an h5 file, then a key to access the target
            DataFrame must be supplied.

    Raises
    ------
    FileNotFoundError
        Input CSV file not found
    DataError
        Metadata could not be used to parse input CSV
    """

    def __init__(
        self,
        dataset_name: str,
        input_file_path: Union[Path, str] = None,
        metadata: Union[dict, Path, str] = None,
        kind: Optional[str] = None,
        init_hash: Optional[Union[str, bytes]] = None,
        h5_key: Optional[str] = None,
    ):
        self.name = dataset_name
        self._history = []
        self._metadata = {}
        self._features = []
        self.df = None
        self.callable_functions = list(
            x for x in dir(Dataset) if callable(getattr(Dataset, x)) and x[0] != "_"
        )

        features_can_be_an_empty_list=False
        
        if metadata is not None:
            if metadata.get("features", None)==[]:
                features_can_be_an_empty_list=True

        if init_hash is None:
            self.sha256 = sha256()
        else:
            if isinstance(init_hash, bytes):
                self.sha256 = sha256(init_hash)
            elif isinstance(init_hash, str):
                self.sha256 = sha256(init_hash.encode("utf-8"))
            else:
                raise ValueError(f"The given argument init_hash was not a bytes array or a string")
        # If file path is a string, make it a Path
        if isinstance(input_file_path, str):
            input_file_path = Path(input_file_path)

        # Allow creation of a blank Dataset object when None is supplied for csv_file_path
        if input_file_path is None:
            self.__update_hash()
            return

        # Load metadata.
        # If str, change to path, check exists, load yaml file defining metadata
        # Can also be a dictionary, which is allowed, or None, relying on defaults.
        metadata = load_dict(metadata, cast_none_to_dict=True)
        
        if kind is not None:
            from .recipes import kinds
            regularised_kind = (
                kind.lower().replace(" ", "").replace("-", "").replace("_", "").replace(".", "")
            )
            if regularised_kind not in kinds:
                raise ValueError(
                    f"Dataset kind given in kind argument was '{kind}',"
                    f" but this was not found in phenonaut. Usable kind arguments are: {kinds.keys()}"
                )
            metadata={**kinds[regularised_kind], **metadata}

        # skip_row_numbers and header_row_number in metadata must be a list
        if isinstance(metadata.get("skip_row_numbers"), int):
            metadata["skip_row_numbers"] = [metadata["skip_row_numbers"]]
        if isinstance(metadata.get("header_row_number"), int):
            metadata["header_row_number"] = [metadata["header_row_number"]]

        if isinstance(input_file_path, Path):
            if input_file_path.suffix == ".h5":
                if "key" not in metadata:
                    raise KeyError("Attempting to load hdf5 file, but no key was given in metadata")
                self.df = pd.read_hdf(input_file_path, h5_key)
            else:
                # Read the CSV file into a pandas DataFrame. It needs to be read in
                # before features are set because the user may be using
                # features_prefix, so we need to examine column names to obtain
                # feature names

                # Pass everything in metadata that can be passed to pd.read_csv
                read_csv_args = {
                    ar: ar_val
                    for ar, ar_val in metadata.items()
                    if ar in signature(pd.read_csv).parameters.keys()
                }
                
                if "sep" not in read_csv_args:
                    # See if we can find a tab in the first line:
                    with open(input_file_path) as myfile:
                        head = next(myfile)
                        if "\t" in head:
                            read_csv_args['sep']="\t"

                self.df = pd.read_csv(input_file_path, **read_csv_args)

            # Some datasets contain column names with spaces after them. Additionally,
            # Columbus column titles often have \xa0 at the end.  Strip this whitespace.
            self.df = self.df.rename(columns=lambda x: x.strip())

        if isinstance(input_file_path, pd.DataFrame):
            self.df = input_file_path.copy()

        # Some documentation and early users rely upon transpose being in metadata.
        # If it is, then we move it to transforms within metadata.
        if "transpose" in metadata:
            if "transforms" not in metadata:
                metadata["transforms"] = []
            metadata["transforms"].extend([("transpose", True)])

        # Perform all transformations
        for t in metadata.get("transforms", []):
            if isinstance(t, str):
                # If string, like "transpose", then make it into a tuple
                t = (t,)

            if isinstance(t, dict):
                if len(t.keys()) != 1:
                    raise ValueError(
                        "Transforms passed as dicts should contain one item, "
                        "where the key is the function name to call. This is "
                        "to stop multiple calls to the same function "
                        "overwriting each other due to the behaviour of "
                        "dictionary keys"
                    )
                t = next(iter(t.items()))

            # At this stage, it should be a list or tuple, or at least
            # something compatible with __getitem__ and with a __len__ method
            # Avoiding using isinstance as much as possible, relying on duck
            # typing.
            # Could be of the form:
            #  ('replace_str',('Var2', 'Pos_', 'Pos-')),
            #  ('replace_str', 3, 'Pos-', 'Pos_'),
            #  ('transpose',)
            if len(t) == 0:
                continue
            elif len(t) == 1:
                func_name, func_args = t[0], ()
            elif len(t) == 2:
                if not isinstance(t[1], (list, tuple)):
                    func_name, func_args = t[0], (t[1],)
                else:
                    func_name, func_args = t
            if len(t) > 2:
                func_name, func_args = t[0], t[1:]

            if func_name not in self.callable_functions:
                raise KeyError(f"'transforms' contains an unknown function: '{func_name}'")
            
            if isinstance(func_args, dict):
                getattr(self, func_name)(**func_args)
            else:
                getattr(self, func_name)(*func_args)

        if "features" in metadata and "initial_features" in metadata:
            raise KeyError(
                "Both 'features' and 'initial_features' found in provided metadata. Please provide only one, if present, 'features' is renamed 'initial_features'"
            )

        # Move features to initial_features
        if "features" in metadata:
            if not isinstance(metadata["features"], (list, tuple, np.ndarray)):
                raise ValueError(
                    f"'features' must be a list, np.nparray, or tuple of strings, it was {type(metadata['features'])}"
                )
            metadata["initial_features"] = metadata["features"]
            del metadata["features"]

        # If no features were given, use features_prefix and features_regex to identify them
        if "initial_features" not in metadata:
            features_prefix = metadata.get("features_prefix")
            features_regex = metadata.get("features_regex")

            if features_prefix is None and features_regex is None:
                print(
                    "No 'features', 'features_prefix', or 'features_regex' given, assuming prefix is 'feat_'"
                )
                features_prefix = ["feat_"]
            if isinstance(features_prefix, str):
                features_prefix = [features_prefix]

            metadata["initial_features"] = []
            if features_prefix is not None:
                # Extract features using features_prefix
                metadata["initial_features"].extend(
                    [
                        column_name
                        for column_name in self.df.columns.to_list()
                        if str(column_name).startswith(tuple(features_prefix))
                    ]
                )

            if features_regex is not None:
                metadata["initial_features"].extend(
                    f
                    for f in list(filter(re.compile(features_regex).match, self.df.columns.values))
                    if f not in metadata["initial_features"]
                )

            if len(metadata["initial_features"]) == 0:
                message = f"No column headers found with specified prefix(es): {features_prefix}, or matching features_regex {features_regex}, first 10 column headers were: {self.df.columns[:10]}"
                raise DataError(message)

        if metadata["initial_features"] is None:
            raise DataError("features cannot be None")
        if len(metadata["initial_features"]) == 0:
            if features_can_be_an_empty_list:
                print(f"Warning, the use of no features was explicitly requested for {dataset_name}")
            else:
                raise DataError(f"features cannot be an empty list, features = {metadata['initial_features']}"
            )

        # Sort features using a natural sort, see:
        # https://blog.codinghorror.com/sorting-for-humans-natural-sort-order/
        # The aim is to make
        # ['f1', 'f10', 'f13', 'f2a', 'f03']
        # sort as:
        # ['f1', 'f2a', 'f03', 'f10', 'f13']
        # ignoring zero padding etc.
        natural_sort_convert_digit_or_str = lambda x: int(x) if x.isdigit() else x.lower()
        metadata["initial_features"] = sorted(
            metadata["initial_features"],
            key=lambda x: [natural_sort_convert_digit_or_str(c) for c in re.split("([0-9]+)", x)],
        )

        # Check all specified features are present in the DataFrame
        missing_features = []
        for feature in metadata["initial_features"]:
            if feature not in self.df.columns:
                print(f"Specified feature ({feature}) not found in columns")
                missing_features.append(feature)
        if len(missing_features) != 0:
            raise ValueError(
                f"The following features were not identified as columns: {missing_features}"
            )

        # At this point in the code, there was a check to ensure that all feature columns contained numeric types.
        # For the CMAP dataset, this took a long time >45 mins using:
        # self.df[metadata["initial_features"]] = self.df[metadata["initial_features"]].apply(
        #     pd.to_numeric, errors="coerce"
        # A supposedly quicker approach tried was:
        # self.df[metadata["initial_features"]] = self.df[metadata["initial_features"]].astype(float)
        # But this took 20 mins.  It was therefore decided to remove the check and rely on Pandas to
        # correctly assign numeric types if it can - this is correct, expected behavior anyway.

        if isinstance(input_file_path, Path):
            self.features = (
                metadata["initial_features"],
                f"{input_file_path=}, {metadata=}, {init_hash=}",
            )

        if isinstance(input_file_path, pd.DataFrame):
            self.features = (
                metadata["initial_features"],
                f"DataFrame passed to Dataset constructor, {metadata=}, {init_hash=}",
            )
        self._metadata = metadata
        self.__update_hash()

    @property
    def data(self):
        """Return self.df[self.features]

        Returns
        -------
        pd.DataFrame
            DataFrame containing only features and index
        """
        return self.df[self.features]

    @property
    def features(self):
        """Return current dataset features

        Returns
        -------
        list
            List of strings containing current features
        """
        return deepcopy(self._features)

    @features.setter
    def features(self, features_and_message_tuple: tuple):
        """Assign new set of feature columns and record a history message

        As well as recording new features and the history message, setting new
        features using this setter causes the hash of the dataset to be updated.

        Parameters
        ----------
        features_and_message_tuple : tuple(list, str)
            Tuple, with the first element being a list of strings documenting
            new features, and the second element is a history message
            documenting what was done to arrive at the new features.
        """

        if not isinstance(features_and_message_tuple, tuple):
            raise AttributeError(
                "Setting features requires a tuple to be supplied of the form:([features], 'history message')"
            )
        if len(features_and_message_tuple) != 2:
            raise AttributeError(
                "Supplied tuple did not have 2 elements. Setting features requires a tuple to be supplied of the form:([features], 'history message')"
            )
        if len(features_and_message_tuple[0]) == 0:
            print(f"Features list is empty, message was {features_and_message_tuple[1]}")
        if not isinstance(features_and_message_tuple[1], str):
            raise AttributeError(
                "Message (second tuple element) must be of type str. Setting features requires a tuple to be supplied of the form:([features], 'history message')"
            )

        self._history.append(TransformationHistory(*features_and_message_tuple))
        self._features = list(features_and_message_tuple[0])
        self.__update_hash()

    def get_history(
        self,
    ) -> list[TransformationHistory]:
        """Get dataset history

        Returns
        -------
        list[TransformationHistory]
            List of TransformationHistory (named tuples) which contain a list
            of features as the first element and then a plain text description
            of what was applied to arrive at those features as the second
            element.
        """
        return self._history

    @property
    def history(self):
        """Get dataset history

        Returns the same as calling .get_history on the dataset

        Returns
        -------
        list[TransformationHistory]
            List of TransformationHistory (named tuples) which contain a list
            of features as the first element and then a plain text description
            of what was applied to arrive at those features as the second
            element.

        """
        return self.get_history()

    @property
    def perturbation_column(self):
        """Return the name of the treatment column

        A treatement is an identifier relating to the peturbation. In many
        cases, it is the unique compound name or identifier.  Many replicates
        may be present, with identifiers like 'DMSO' etc.

        Returns
        -------
        String
            Column name of dataframe containing the treatment.
        """
        perturbation_column = self._metadata.get("perturbation_column", None)
        if perturbation_column is None:
            print(
                f"Warning: Attempted to get dataset ({self.name=}) perturbation_column, but it is None"
            )
        return perturbation_column

    def __str__(self):
        return f"Phenonaut.Dataset:'{self.name}', shape={self.df.shape}, n_features={len(self.features)}"

    def __repr__(self):
        return f"Phenonaut.Dataset:'{self.name}', shape={self.df.shape}, n_features={len(self.features)}"

    @perturbation_column.setter
    def perturbation_column(self, perturbation_column: str):
        """Assign new treatment column name

        A treatement is an identifier relating to the peturbation. In many
        cases, it is the unique compound name or identifier.  Many replicates
        may be present, with identifiers like 'DMSO' etc.

        Parameters
        ----------
        perturbation_column : str
            New column name which contains the peturbation IDs.
        """
        self._metadata["perturbation_column"] = perturbation_column
        self.features = (
            self.features,
            f"Updated perturbation column, perturbation column is now {perturbation_column}",
        )

    def get_unique_perturbations(
        self,
    ):
        if self.perturbation_column is None:
            return None
        return [t for t in self.df[self.perturbation_column].unique() if not pd.isna(t) and t != ""]

    def get_feature_ranges(self, pad_by_percent: Union[float, int]) -> tuple:
        """Get the ranges of feature columns

        Parameters
        ----------
        pad_by_percent : Union[float, int]
            Optionally, pad the ranges by a percent value

        Returns
        -------
        tuple
            Returns tuple of tuples with shape (features, 2), for example,
            with two features it would return:
            ((min_f1, max_f1), (min_f2, max_f2))
        """
        feature_min_max = []
        for feature in self.features:
            feature_min = self.df[feature].min()
            feature_max = self.df[feature].max()
            feature_range = feature_max - feature_min
            feature_min_max.append(
                (
                    feature_min - ((feature_range / 100) * float(pad_by_percent)),
                    feature_max + ((feature_range / 100) * float(pad_by_percent)),
                )
            )
        return tuple(feature_min_max)

    def get_non_feature_columns(self) -> list:
        """Get columns which are not features

        Returns
        -------
        list[str]
            Returns list of Dataset columns which are not currently features.
        """
        return list(set(self.df.columns.unique().values) - set(self.features))

    def drop_columns(
        self, column_labels: Union[str, list[str], tuple[str]], reason: Optional[str] = None
    ) -> None:
        """Drop columns inplace, update features if needed and set new history.

        Intelligently drop columns from the dataset (inplace). If any of those
        columns were listed as features, then remove them from the features list
        and set a new new history.  Updating features and new history only
        happens if it needs to (removed column was a feature). Updaing of
        features will cause hash update.

        Parameters
        ----------
        column_labels : Union[str, list[str], tuple[str]]
            List of column labels which should be removed. Can also be a str to
            remove just one column.
        reason : Optional[str]
            A reason may be given for dropping the column. If not None and the
            column was a feature, then this reason is recorded along with the
            history. If None, and the column was a feature, then the history
            entry will state:
            "Droped columns ({column_labels})"
            Where column_labels contains the dropped columns.
            If reason is not None and the column is a feature, then the history
            entry will state:
            "Droped columns ({column_labels}), reason:{reason}"
            where {reason} is the given reason. Has no effect if the dropped
            column is not a feature, or the list of dropped columns do not
            contain a feature. By default None.
        """

        if isinstance(column_labels, str):
            column_labels = [column_labels]
        features_removed = set(self.features).intersection(set(column_labels))

        self.df.drop(columns=column_labels, inplace=True)

        reason_message = (
            f", reason{reason}" if reason is not None and len(features_removed) > 0 else ""
        )

        if len(features_removed) > 0:
            new_features = [feat for feat in self.features if feat not in features_removed]
            self.features = new_features, f"Droped columns ({column_labels}) {reason_message}"

    def filter_inplace(self, query: str) -> None:
        """Apply a pandas query style filter, keeping all that pass

        Parameters
        ----------
        query : str
            Pandas style query which when applied to rows, will keep all those
            which return True.
        """
        self.df = self.df.loc[self.df.query(query).index]
        self.features = self.features, f"Filtered in place, query={query}"

    def subtract_median(self, query: str) -> None:
        """Subtract the median of rows identified in the query from features

        Useful function for working with imaging plates, where conventionally,
        median DMSO control wells on a per-plate basis are subtracted.

        Parameters
        ----------
        query : str
            Pandas style query to retrieve rows from which medians are
            calculated.
        """
        self.df[self.features] = self.data - self.df.query(query)[self.features].median()
        self.features = self.features, f"Subtracted median of rows matching query: {query}"

    def subtract_mean(self, query: str) -> None:
        """Subtract the mean of rows identified in the query from features

        Useful function for normalising to controls.

        Parameters
        ----------
        query : str
            Pandas style query to retrieve rows from which means are calculated.
        """
        self.df[self.features] = self.data - self.df.query(query)[self.features].mean()
        self.features = self.features, f"Subtracted mean of rows matching query: {query}"

    def divide_median(self, query: str) -> None:
        """Divide dataset features by the median of rows identified in the query

        Useful function for normalising to controls.

        Parameters
        ----------
        query : str
            Pandas style query to retrieve rows from which medians are
            calculated.
        """
        self.df[self.features] = self.data / self.df.query(query)[self.features].median()
        self.features = self.features, f"Divided by median of rows matching query: {query}"

    def divide_mean(self, query: str) -> None:
        """Divide dataset features by the mean of rows identified in the query

        Useful function for normalising to controls.

        Parameters
        ----------
        query : str
            Pandas style query to retrieve rows from which means are calculated.
        """
        self.df[self.features] = self.data / self.df.query(query)[self.features].mean()
        self.features = self.features, f"Divided by mean of rows matching query: {query}"

    def distance_df(
        self,
        candidate_dataset: "Dataset",
        metric: Union[str, Callable] = "euclidean",
        return_best_n_indexes_and_score: Optional[int] = None,
        lower_is_better=True,
    ) -> pd.DataFrame:
        """Generate a distance DataFrame

        Distance DataFrames allow simple generation of pd.DataFrames where the
        index take the form of perturbations and the columns other
        perturbations. The values at the intersections are therfore the
        distances between these perturbations in feature space. Many different
        metrics both inbuilt and custom/user defined may be used.

        Parameters
        ----------
        candidate_dataset : Dataset
            The dataset to which the query (this) should be compared.
        metric : Union[str, Callable], optional
            Metric which should be used for the distance calculation.  May be a
            simple string understood by scipy.spatial.distance.cdist, or a
            callable, like a function or lambda accepting two vectors
            representing query and candidate features. By default "euclidean".
        return_best_n_indexes_and_score : Optional[int], optional
            If an integer is given, then just that number of best pairs/measures
            are returned. By default None.
        lower_is_better : bool, optional
            If using the above 'return_best_n_indexes_and_score' then it needs
            to be flagged if lower is better (default), or higher is better.
            By default True

        Returns
        -------
        Union [pd.DataFrame, tuple(tuple(int, int), float)]
            Returns a distance Dataframe, unless
            'return_best_n_indexes_and_score' is an int, in which case a list of
            the top scoring pairs are returned in the form of a nested tuple:
            ((from, to), score)

        Raises
        ------
        ValueError
            Error raised if this Dataset and the given candidate Dataset do not
            share common features.
        """
        if not all([f in candidate_dataset.features for f in self.features]):
            raise ValueError("Datasets do not share common features")

        dist_mat = cdist(self.data, candidate_dataset.df[self.features], metric=metric)

        if return_best_n_indexes_and_score is None:
            return pd.DataFrame(
                data=dist_mat, index=self.df.index, columns=candidate_dataset.df.index.to_list()
            )

        else:
            if lower_is_better:
                top_n = list(
                    map(
                        lambda x: np.unravel_index(x, dist_mat.shape),
                        np.argsort(dist_mat, axis=None)[:return_best_n_indexes_and_score],
                    )
                )
            else:
                top_n = list(
                    map(
                        lambda x: np.unravel_index(x, dist_mat.shape),
                        np.argsort(dist_mat, axis=None)[-return_best_n_indexes_and_score:],
                    )
                )[::-1]
        return [(loc, dist_mat[loc[0], loc[1]]) for loc in top_n]

    def get_ds_from_query(
        self,
        name: str,
        query: str,
    ):
        """Make a new Dataset object from a pandas style query.

        Parameters
        ----------
        name : str
            Name of new dataset
        query : str
            Pandas style query from which all rows returning true will be
            included into the new PhenonautGenericData set object.

        Returns
        -------
        Dataset
            New dataset created from query
        """
        new_ds = Dataset(name, input_file_path=None, metadata=None)
        new_ds.df = self.df.query(query)
        new_ds.features = (self.features, f"New dataset from query: {query}")
        new_ds._metadata = self._metadata
        return new_ds

    def copy(self):
        """Return a deep copy of the Dataset object

        Returns
        -------
        PhenonautData
            Copy of the input object.
        """
        tmp_hash_object = self.sha256.copy()
        self.sha256 = None
        new_ds = deepcopy(self)
        self.sha256 = tmp_hash_object
        new_ds.sha256 = self.sha256.copy()
        return new_ds

    def filter_columns_with_prefix(self, column_prefix: Union[str, list], keep: bool = False):
        """Filter columns based on prefix

        Parameters
        ----------
        column_prefix : Union[str, list]
            Prefix for columns as a string, or alternatively, a list of string
            prefixes
        keep : bool, optional
            If true, only columns matching the prefix are kept, if false, these
            columns are removed, by default False
        """
        column_names = []
        if isinstance(column_prefix, list):
            column_prefix = tuple(column_prefix)
        column_names = [c for c in self.df.columns.values if c.startswith(column_prefix)]
        self.filter_columns(column_names, keep=keep)

    def filter_columns(self, column_names: list, keep=True, regex=False):
        """Filter dataframe columns

        Parameters
        ----------
        column_names : list
            Column names
        keep : bool, optional
            Keep columns listed in column_names, if false, then the opposite
            happens and these columns are removed, by default True.
        """

        if regex:
            new_column_names = []
            for pattern in column_names:
                for match in list(filter(re.compile(pattern).match, self.df.columns.values)):
                    if match not in new_column_names:
                        new_column_names.append(match)
            column_names = new_column_names

        if len(column_names) == 0:
            return
        if keep:
            columns_for_removal = set(self.df.columns.values) - set(column_names)
            features_to_remove = set(self.features).intersection(columns_for_removal)
            if len(features_to_remove) > 0:
                self.features = (
                    [f for f in self.features if f not in features_to_remove],
                    f"Features removed due to filtering of columns, columns kept were {column_names=}, resulting in removal of the following columns  {columns_for_removal=}, which removed the following features {features_to_remove}",
                )
            self.df = self.df[column_names]

        else:
            features_to_remove = set(self.features).intersection(column_names)
            if len(features_to_remove) > 0:
                self.features = (
                    [f for f in self.features if f not in features_to_remove],
                    f"Features removed due to filtering of columns, columns removed were {column_names=}, which removed the following features {features_to_remove}",
                )
            self.df = self.df.drop(column_names, axis=1)

    def filter_rows(self, query_column: str, values: Union[list, str], keep: bool = True):
        """Filter dataframe rows

        Parameters
        ----------
        query_column : str
            Column name which is being filtered on
        values : Union[list, str]
            List or string of values to be filtered on
        keep : bool, optional
            If true, then only rows containing listed values in query column
            are kept. If this argument is false, then the opposite occurs, and
            the rows matching are discarded, by default True

        """

        if not isinstance(values, list):
            values = [values]

        if not query_column in self.df.columns:
            raise KeyError(
                f"Could not find {query_column} in dataframe columns, columns are: {self.df.columns.values}"
            )
        if keep:
            self.df = self.df.query(f"{query_column} in @values")
        else:
            self.df = self.df.query(f"{query_column} not in @values")
        self.features = (self.features, f"Filtered rows, {query_column=}, {values=}")

    def rename_column(self, from_column_name: str, to_column_name: str):
        """Rename a single dataset column

        Parameters
        ----------
        from_column_name : str
            Name of column to rename
        to_column_name : str
            New column name
        """
        self.rename_columns({from_column_name: to_column_name})

    def rename_columns(self, from_to: dict):
        """Rename multiple columns

        Parameters
        ----------
        from_to : dict
            Dictionary of the form {'old_name':'new_name'}
        """
        self.df.rename(columns=from_to, inplace=True)
        if len(self._history) != 0:
            self.features = self.features, f"Renamed columns: {from_to}"

    def df_to_csv(self, output_path: Union[Path, str], **kwargs):
        """Write DataFrame to CSV

        Convenience function to write the underlying DataFrame to a CSV.  Additional
        arguments will be passed to the Pandas.DataFrame.to_csv function.

        Parameters
        ----------
        output_path : Union[Path, str]
            Target output file
        """
        if isinstance(output_path, str):
            output_path = Path(output_path)
        if "index" not in kwargs:
            kwargs["index"] = False
        self.df.to_csv(output_path, **kwargs)

    def df_to_multiple_csvs(
        self,
        split_by_column: str,
        output_dir: Optional[Union[str, Path]] = None,
        file_prefix: str = "",
        file_suffix="",
        file_extension=".csv",
        **kwargs,
    ):
        """Wite multiple CSV files from a dataset DataFrame.

        In the case where one output CSV is required per plate, then splitting
        the underlying DataFrame on something like a PlateID serves the purpose
        of generating one output CSV file per plate.  This can be achieved with
        this function and providing the column to split on.

        Parameters
        ----------
        split_by_column : str
            Column containing unique values within a split output CSV file
        output_dir : Optional[Union[str, Path]], optional
            Target output directory for split CSV files, by default None
        file_prefix : str, optional
            Prefix for split CSV files, by default ""
        file_suffix : str, optional
            Suffix for split CSV files, by default ""
        file_extension : str, optional
            File extension for split CSV files, by default ".csv"
        """

        if isinstance(output_dir, str):
            output_dir = Path(output_dir)
            if not output_dir.exists():
                output_dir.mkdir(parents=True)

        if "index" not in kwargs:
            kwargs["index"] = False

            splits = self.df[split_by_column].unique()
            for split in splits:
                if split == np.nan:
                    continue
                split_output_path = (
                    output_dir / f"{file_prefix}{split}{file_suffix}{file_extension}"
                )
                self.df.query(f"{split_by_column} == '{split}'").to_csv(split_output_path, **kwargs)

    def add_well_id(
        self,
        numerical_column_name: str = "COLUMN",
        numerical_row_name: str = "ROW",
        plate_type: int = 384,
        new_well_column_name: str = "Well",
        add_empty_wells: bool = False,
        plate_barcode_column: str = None,
        no_sort: bool = False,
    ):
        """Add standard well IDs - such as A1, A2, etc.

        If a dataset contains numerical row and column names, then they may be
        translated into standard letter-number well IDs.

        Parameters
        ----------
        numerical_column_name : str, optional
            Name of column containing numeric column number, by default
            "COLUMN".
        numerical_row_name : str, optional
            Name of column containing numeric column number, by default "ROW".
        plate_type : int, optional
            Plate type - note, at present, only 384 well plate format is
            supported, by default 384.
        new_well_column_name : str, optional
            Name of new column containing letter-number well ID, by default
            "Well".
        add_empty_wells : bool, optional
            Should all wells from a plate be inserted, even when missing from
            the data, by default False.
        plate_barcode_column : str, optional
            Multiple plates may be in a dataset, this column contains their
            unique ID, by default None.
        no_sort : bool, optional
            Do not resort the dataset by well ID, by default False.
        """
        if plate_type not in [384]:
            message = f"Attempted to assign alphanumeric wells (like B01) to requested {plate_type} well plate.  {plate_type} well plates not implemented."
            raise ValueError(message)

        MAX_ROWS = None
        MAX_COLUMNS = None
        row_map = None
        column_map = None
        if plate_type == 384:
            MAX_ROWS = 16
            MAX_COLUMNS = 24
            row_map = {n: chr(n + 64) for n in range(1, MAX_ROWS + 1)}
            column_map = {n: f"{n}".zfill(2) for n in range(1, MAX_COLUMNS + 1)}

        # Make sure to cast to a list, otherwise it is consumed upon use/iteration
        all_possible_well_ids = list(itertools_product(row_map.keys(), column_map.keys()))
        numerical_column = self.df[numerical_column_name].astype(int)
        numerical_row = self.df[numerical_row_name].astype(int)
        if numerical_column.max() > MAX_COLUMNS:
            raise DataError(
                f"Column value {numerical_column.max()} is greater than maximum for {plate_type} well plate of {MAX_COLUMNS}"
            )
        if numerical_row.max() > MAX_ROWS:
            raise DataError(
                f"Row value {numerical_row.max()} is greater than maximum for {plate_type} well plate of {MAX_ROWS}"
            )

        self.df[new_well_column_name] = numerical_row.apply(lambda x: row_map[x]).astype(
            str
        ) + numerical_column.apply(lambda x: column_map[x]).astype(str)

        if add_empty_wells:
            if plate_barcode_column is None:
                all_well_ids = list(zip(numerical_row, numerical_column))
                missing_wells = list((set(all_possible_well_ids) - set(all_well_ids)))
                if len(missing_wells) > 0:
                    additional_data = pd.DataFrame.from_dict(
                        {
                            numerical_row_name: [w[0] for w in missing_wells],
                            numerical_column_name: [w[1] for w in missing_wells],
                            new_well_column_name: [
                                row_map[w[0]] + column_map[w[1]] for w in missing_wells
                            ],
                        }
                    )
                    self.df = pd.concat([self.df, additional_data], axis=0, ignore_index=True)
                    if not no_sort:
                        self.df = self.df.sort_values([new_well_column_name])
            else:
                for plate in self.df[plate_barcode_column].unique():
                    if plate is np.nan:
                        continue
                    plate_df = self.df.query(f"{plate_barcode_column} == '{plate}'")
                    all_well_ids = list(
                        zip(
                            plate_df[numerical_row_name],
                            plate_df[numerical_column_name],
                        )
                    )
                    missing_wells = [
                        (w[0], w[1], plate)
                        for w in (set(all_possible_well_ids) - set(all_well_ids))
                    ]

                    if len(missing_wells) > 0:
                        additional_wells = pd.DataFrame.from_dict(
                            {
                                numerical_row_name: [w[0] for w in missing_wells],
                                numerical_column_name: [w[1] for w in missing_wells],
                                new_well_column_name: [
                                    row_map[w[0]] + column_map[w[1]] for w in missing_wells
                                ],
                                plate_barcode_column: [w[2] for w in missing_wells],
                            }
                        )
                        self.df = pd.concat([self.df, additional_wells], axis=0, ignore_index=True)
                        if not no_sort:
                            self.df = self.df.sort_values(
                                [plate_barcode_column, new_well_column_name]
                            )
        self.features = (
            self.features,
            f"Added well IDs, {numerical_column_name=}, {numerical_row_name=}, {plate_type=}, {new_well_column_name}, {add_empty_wells=}, {plate_barcode_column=}, {no_sort=}",
        )

    def subtract_median_perturbation(
        self,
        perturbation_label: str,
        per_column_name: Optional[str] = None,
        new_features_prefix: str = "SMP_",
    ):
        """Subtract the median perturbation from all features

        Useful for normalisation within a well/plate format. The median feature
        may be identified through the per_column_name variable, and perturbation
        label. Newly generated features may have their prefixes controled via
        the new_features_prefix argument.

        Parameters
        ----------
        perturbation_label : str
            The perturbation label which should be used to calculate the median
        per_column_name : Optional[str], optional
            The perturbation column name.  This is optional and can be None, as
            the Dataset may already have perturbation column set. By
            default, None.
        new_features_prefix : str
            Prefix for new features, each with the median perturbation
            subtracted. By default 'SMP\_' (for subtracted median perturbation).
        """
        new_features = [f"SMP_{feat}" for feat in self.features]
        if per_column_name is None:
            self.df[new_features] = self.data - self.df.median(axis=0)
        else:
            for unique_plate in self.df[per_column_name].unique():
                medians = self.df.loc[
                    self.df[per_column_name] == unique_plate, self.features
                ].median(axis=0)
                self.df.loc[
                    self.df[self.df[per_column_name] == unique_plate].index, new_features
                ] = (
                    self.df.loc[
                        self.df[self.df[per_column_name] == unique_plate].index, self.features
                    ]
                    - medians
                ).values
        self.features = (
            new_features,
            f"Subtracted median perturbation {perturbation_label=}, {per_column_name=}",
        )

    def transpose(self, reset_index: bool = True, new_header_column: Optional[int] = 0):
        """Transpose internal DataFrame"""

        self.df = self.df.transpose()
        if reset_index:
            self.df = self.df.reset_index()

        if new_header_column is not None:
            self.df.columns = self.df.iloc[new_header_column]
            self.df.drop(index=new_header_column, inplace=True)

    def replace_str(self, column: Union[str, int], pat: str, repl: str):
        """Replace a string present in a column

        Parameters
        ----------
        column : Union[str, int]
            Name of the column(could be a feature), within which to search and replace instances of the string specified in the 'pat' argument.
        pat : str
            The patter, (non-regex), just query substring to find and replace.
        repl : str
            Replacement text for the substring identified in the 'pat' argument.
        """
        if isinstance(column, int):
            print(f"replacing in {column}, {self.df.columns.tolist()}")
            column = self.df.columns.tolist()[column]
        self.df[column] = self.df[column].str.replace(pat, repl)

    def split_column(self, column: Union[str, int], pat: str, new_columns: list[str]):
        """Split a column on a delimiter

        If a column named 'data' contained:

        === ===============
        idx data
        === ===============
        1   A2_CPD1_Plate1
        === ===============

        Then calling:

        .. code-block:: python

            split_column('data', '_', ['WellID', 'CpdID', 'PlateID'])

        Would introduce the following new columns into the dataframe:

        === ====== ===== =======
        idx WellID CpdID PlateID
        === ====== ===== =======
        1   A2     CPD1  Plate1
        === ====== ===== =======

        Parameters
        ----------
        column : Union[str, int]
            Name of column to split, or the index of the column.
        pat : str
            Pattern (non-regex), usually a delimiter to split on.
        new_columns : list[str]
            List of new column names. Should be the correct size to absorb all
            produced splits.

        Raises
        ------
        DataError
            Inconsistent number of splits produced when splitting the column.
        ValueError
            Incorrect number of new column names given in new_columns.
        """
        if isinstance(column, int):
            column = self.df.columns.tolist()[column]
        pat_count = self.df[column].str.count("_").unique().tolist()
        if len(pat_count) != 1:
            different_count_indexes = [
                self.df.loc[self.df.loc[self.df[column].str.count(pat) == c].index[0], column]
                for c in pat_count
            ]

            raise DataError(
                f"Attempting to split {column} column on '{pat}', "
                f"but column does not contain a consistent count of '{pat}',"
                f" contained: {pat_count} instances, from {column} values such as:\n"
                f"{different_count_indexes}"
            )
        if len(new_columns) != pat_count[0] + 1:
            raise ValueError(
                f"'new_columns' argument to split_column contains "
                f"{len(new_columns)} elements, but splitting on '{pat}' "
                f"produces {pat_count[0]} new columns"
            )

        self.df[new_columns] = self.df[column].str.split(pat, expand=True)

    def pivot(self, feature_names_column: str, values_column: str):
        self.df = self.df.pivot_table(
            values=values_column,
            index=[c for c in self.df.columns if c not in [feature_names_column, values_column]],
            columns=feature_names_column,
        )
        self.df.reset_index(inplace=True)
        self.df.columns.name = ""

    def new_aggregated_dataset(
        self,
        identifier_columns: list[str],
        new_dataset_name: str = "Merged rows dataset",
        transformation_lookup: dict[str, Union[Callable, str]] = None,
        tranformation_lookup_default_value: Union[str, Callable] = "mean",
    ):
        """Merge dataset rows and make a new dataframe

        If we have a pd.DataFrame containing data derived from 2 fields of view
        from a microscopy image, a sensible approach is averaging features.
        If we have the DataFrame below, we may merge FOV 1 and FOV 2, taking the
        mean of all features.  As strings such as filenames should be kept, they
        are concatenated together, separated by a comma, unless the strings are
        the same, in which case just one is used.

        Here we test a df as follows:

        === ======= ======= ======= ======= ======= =========== ===
        ROW COLUMN  BARCODE feat_1  feat_2  feat_3  filename    FOV
        === ======= ======= ======= ======= ======= =========== ===
        1   1       Plate1  1.2	    1.2	    1.3	    fileA.png   1
        1   1       Plate1  1.3	    1.4	    1.5	    FileB.png   2
        1   1       Plate2  5.2	    5.1	    5	    FileC.png   1
        1   1       Plate2  6.2	    6.1	    6.8	    FileD.png   2
        1   2       Plate1  0.1	    0.2	    0.3	    fileE.png   1
        1   2       Plate1  0.2	    0.2	    0.38    FileF.png   2
        === ======= ======= ======= ======= ======= =========== ===


        Merging produces:

        ==== ====== ======== ======= ======= ====== =================== =====
        ROW  COLUMN BARCODE  feat_1  feat_2  feat_3            filename   FOV
        ==== ====== ======== ======= ======= ====== =================== =====
        1       1   Plate1    1.25     1.3    1.40  fileA.png,FileB.png   1.5
        1       1   Plate2    5.70     5.6    5.90  FileC.png,FileD.png   1.5
        1       2   Plate1    0.15     0.2    0.34  FileF.png,fileE.png   1.5
        ==== ====== ======== ======= ======= ====== =================== =====

        Note that the FOV column has also been averaged.

        Parameters
        ----------
        identifier_columns : list[str]
            If a biochemical assay evaluated through imaging is identified by a
            row, column, and barcode (for the plate) but multiple images taken
            from a well, then these multiple fields of view can be merged,
            creating averaged features.
        new_dataset_name : str, optional
            Name for the new Dataset, by default "Merged rows dataset"
        transformation_lookup : dict[str,Union[Callable, str]]
            Dictionary mapping data types to aggregations. When None, it is as
            if the dictionary:
            {np.dtype("O"): lambda x: ",".join([f"{item}" for item in set(x)])}
            was provided, concatenating strings together (separated by a comma)
            if they are different and just using one if they are the same across
            rows. If a type not present in the dictionary is encountered (such
            as int, or float) in the above example, then the default specified
            by transformation_lookup_default_value is returned.
            By default, None.
        tranformation_lookup_default_value : Union[str, Callable]
            Transformation to apply if the data type is not found in the
            transformation_lookup_dictionary, can be a callable or string to
            pandas defined string to function shortcut mappings.
            By default "mean".

        Returns
        -------
        Dataset
            Dataset with samples merged.
        """

        if transformation_lookup is None:
            transformation_lookup = {
                np.dtype("O"): lambda x: ",".join([f"{item}" for item in set(x)])
            }
        dtypes_dict = {n: dt for n, dt in zip(self.df.columns, self.df.dtypes)}
        transforms = {
            c: transformation_lookup.get(dt, tranformation_lookup_default_value)
            for c, dt in zip(self.df.columns, self.df.dtypes)
        }
        merged_df = self.df.copy().groupby(identifier_columns, as_index=False).aggregate(transforms)
        
        for id_col_name in identifier_columns:
            merged_df[id_col_name] = merged_df[id_col_name].astype(dtypes_dict[id_col_name])
        tmp_metadata = self._metadata.copy()
        if "features_prefix" in tmp_metadata:
            del tmp_metadata["features_prefix"]
        if "initial_features" in tmp_metadata:
            del tmp_metadata["initial_features"]
        tmp_metadata['features']=self.features
        
        
        new_ds = Dataset(dataset_name=new_dataset_name, input_file_path=merged_df, metadata=tmp_metadata)
        new_ds._history = self.get_history()
        new_ds.features = (
            self.features,
            f"Merged rows of the dataset, using columns {identifier_columns}",
        )
        return new_ds

    def __update_hash(
        self,
    ):
        self.sha256.update(self.name.encode("utf-8"))
        self.sha256.update(
            ";".join(f"{','.join(feat)}:{t_hist}" for feat, t_hist in self.get_history()).encode(
                "utf-8"
            )
        )
        self.sha256.update("".join(self._metadata).encode("utf-8"))
        self.sha256.update(",".join(self.features).encode("utf-8"))
        if self.df is not None:
            self.sha256.update(self.df.values.tobytes())
            self.sha256.update(self.df.columns.values.tobytes())



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from curses import meta
import pytest
import phenonaut
from iris import *
from phenonaut.data.dataset import Dataset
import pandas as pd
import numpy as np
import tempfile
from io import StringIO


from phenonaut.phenonaut import Phenonaut


def test_possible_dataset_combinations(phenonaut_object_iris_4_views):
    all_possible_views = phenonaut_object_iris_4_views.get_dataset_combinations()
    assert all_possible_views == (
        ("Iris_view1",),
        ("Iris_view2",),
        ("Iris_view3",),
        ("Iris_view4",),
        ("Iris_view1", "Iris_view2"),
        ("Iris_view1", "Iris_view3"),
        ("Iris_view1", "Iris_view4"),
        ("Iris_view2", "Iris_view3"),
        ("Iris_view2", "Iris_view4"),
        ("Iris_view3", "Iris_view4"),
        ("Iris_view1", "Iris_view2", "Iris_view3"),
        ("Iris_view1", "Iris_view2", "Iris_view4"),
        ("Iris_view1", "Iris_view3", "Iris_view4"),
        ("Iris_view2", "Iris_view3", "Iris_view4"),
        ("Iris_view1", "Iris_view2", "Iris_view3", "Iris_view4"),
    )


def test_possible_dataset_combinations_indexes(phenonaut_object_iris_4_views):
    all_possible_views = phenonaut_object_iris_4_views.get_dataset_combinations(return_indexes=True)
    assert all_possible_views == (
        (0,),
        (1,),
        (2,),
        (3,),
        (0, 1),
        (0, 2),
        (0, 3),
        (1, 2),
        (1, 3),
        (2, 3),
        (0, 1, 2),
        (0, 1, 3),
        (0, 2, 3),
        (1, 2, 3),
        (0, 1, 2, 3),
    )


def test_get_dataset_index_from_name(phenonaut_object_iris_4_views):
    assert phenonaut_object_iris_4_views.get_dataset_index_from_name(
        ["Iris_view4", "Iris_view3", "Iris_view2", "Iris_view1"]
    ) == [3, 2, 1, 0]


def test_phenonaut_obj_from_df():
    df = pd.DataFrame({"id": [9, 8, 7], "A": [1, 2, 3], "B": [4, 5, 6]})
    another_df = pd.DataFrame({"id2": [9, 8, 7], "A2": [1, 2, 3], "B": [4, 5, 6]})
    ph0 = phenonaut.Phenonaut(df, metadata={"features": ["B"]}, dataframe_name="Bob")
    assert ph0.ds.features == ["B"]
    assert ph0.ds.name == "Bob"
    ph0 = phenonaut.Phenonaut(df, metadata={"features_prefix": ["B"]})
    assert ph0.ds.features == ["B"]
    ph0 = phenonaut.Phenonaut(df, metadata={"features_prefix": ""})
    assert set(ph0.ds.features) == set(["B", "A", "id"])
    ph0 = phenonaut.Phenonaut([df, another_df], metadata={"features_prefix": ""})
    assert set(ph0.datasets[0].features) == set(["B", "A", "id"])
    assert set(ph0.datasets[1].features) == set(["B", "A2", "id2"])
    ph0 = phenonaut.Phenonaut([df, another_df], metadata={"features": ["B"]})
    assert set(ph0.datasets[0].features) == set(["B"])
    assert set(ph0.datasets[1].features) == set(["B"])
    ph0 = phenonaut.Phenonaut(
        [df, another_df],
        metadata=[{"features": ["B"]}, {"features_prefix": ""}],
        dataframe_name=["Bob", "Robert"],
    )
    assert set(ph0.datasets[0].features) == set(["B"])
    assert set(ph0.datasets[1].features) == set(["A2", "B", "id2"])
    assert ph0.datasets[0].name == "Bob"
    assert ph0.datasets[-1].name == "Robert"


def test_aggregate_dataset_in_ph0_obj():
    """Here we test a df as follows:
    ROW	COLUMN	BARCODE	feat_1	feat_2	feat_3	filename	FOV
    1	1	    Plate1	1.2	    1.2	    1.3	    fileA.png	1
    1	1	    Plate1	1.3	    1.4	    1.5	    FileB.png	2
    1	1	    Plate2	5.2	    5.1	    5	    FileC.png	1
    1	1	    Plate2	6.2	    6.1	    6.8	    FileD.png	2
    1	2	    Plate1	0.1	    0.2	    0.3	    fileE.png	1
    1	2	    Plate1	0.2	    0.2	    0.38    FileF.png	2

    As there are multiple fields of view, we merge on common ROW, COLUMN and BARCODE fields.
    Merging uses np.mean, but may be any callable which returns a single value

    """

    df = pd.DataFrame(
        {
            "ROW": [1, 1, 1, 1, 1, 1],
            "COLUMN": [1, 1, 1, 1, 2, 2],
            "BARCODE": ["Plate1", "Plate1", "Plate2", "Plate2", "Plate1", "Plate1"],
            "feat_1": [1.2, 1.3, 5.2, 6.2, 0.1, 0.2],
            "feat_2": [1.2, 1.4, 5.1, 6.1, 0.2, 0.2],
            "feat_3": [1.3, 1.5, 5, 6.8, 0.3, 0.38],
            "filename": [
                "fileA.png",
                "FileB.png",
                "FileC.png",
                "FileD.png",
                "fileE.png",
                "FileF.png",
            ],
            "FOV": [1, 2, 1, 2, 1, 2],
        }
    )

    phe = phenonaut.Phenonaut(df, metadata={"features_prefix": "feat_"})
    phe.aggregate_dataset(["ROW", "COLUMN", "BARCODE"])
    assert len(phe.datasets) == 2
    assert set(phe.ds.features) == set(["feat_1", "feat_2", "feat_3"])
    assert abs(phe.ds.df["feat_1"][0] - 1.25 < 0.00001)
    assert np.abs(phe.ds.df["feat_1"][1] - 5.70 < 0.00001)
    assert np.abs(phe.ds.df["feat_1"][2] - 0.15 < 0.00001)
    assert np.abs(phe.ds.df["feat_2"][0] - 1.3 < 0.00001)
    assert np.abs(phe.ds.df["feat_2"][1] - 5.6 < 0.00001)
    assert np.abs(phe.ds.df["feat_2"][2] - 0.2 < 0.00001)
    assert np.abs(phe.ds.df["feat_3"][0] - 1.4 < 0.00001)
    assert np.abs(phe.ds.df["feat_3"][1] - 5.9 < 0.00001)
    assert np.abs(phe.ds.df["feat_3"][2] - 0.34 < 0.00001)


def test_aggregate_dataset():
    """Here we test a df as follows:
    ROW	COLUMN	BARCODE	feat_1	feat_2	feat_3	filename	FOV
    1	1	    Plate1	1.2	    1.2	    1.3	    fileA.png	1
    1	1	    Plate1	1.3	    1.4	    1.5	    FileB.png	2
    1	1	    Plate2	5.2	    5.1	    5	    FileC.png	1
    1	1	    Plate2	6.2	    6.1	    6.8	    FileD.png	2
    1	2	    Plate1	0.1	    0.2	    0.3	    fileE.png	1
    1	2	    Plate1	0.2	    0.2	    0.38    FileF.png	2

    As there are multiple fields of view, we merge on common ROW, COLUMN and BARCODE fields.
    Merging uses np.mean, but may be any callable which returns a single value

    """
    df = pd.DataFrame(
        {
            "ROW": [1, 1, 1, 1, 1, 1],
            "COLUMN": [1, 1, 1, 1, 2, 2],
            "BARCODE": ["Plate1", "Plate1", "Plate2", "Plate2", "Plate1", "Plate1"],
            "feat_1": [1.2, 1.3, 5.2, 6.2, 0.1, 0.2],
            "feat_2": [1.2, 1.4, 5.1, 6.1, 0.2, 0.2],
            "feat_3": [1.3, 1.5, 5, 6.8, 0.3, 0.38],
            "filename": [
                "fileA.png",
                "FileB.png",
                "FileC.png",
                "FileD.png",
                "fileE.png",
                "FileF.png",
            ],
            "FOV": [1, 2, 1, 2, 1, 2],
        }
    )

    phe = phenonaut.Phenonaut([df, df])
    phe.aggregate_dataset(
        ["ROW", "COLUMN", "BARCODE"], datasets=[0, 1], new_names_or_prefix=["M1", "M2"]
    )
    assert len(phe.datasets) == 4
    phe = phenonaut.Phenonaut([df, df])
    phe.aggregate_dataset(
        ["ROW", "COLUMN", "BARCODE"],
        datasets=[0, 1],
        new_names_or_prefix=["M1", "M2"],
        inplace=True,
    )
    assert len(phe.datasets) == 2
    assert set(phe.ds.features) == set(["feat_1", "feat_2", "feat_3"])
    assert abs(phe.ds.df["feat_1"][0] - 1.25 < 0.00001)
    assert np.abs(phe.ds.df["feat_1"][1] - 5.70 < 0.00001)
    assert np.abs(phe.ds.df["feat_1"][2] - 0.15 < 0.00001)
    assert np.abs(phe.ds.df["feat_2"][0] - 1.3 < 0.00001)
    assert np.abs(phe.ds.df["feat_2"][1] - 5.6 < 0.00001)
    assert np.abs(phe.ds.df["feat_2"][2] - 0.2 < 0.00001)
    assert np.abs(phe.ds.df["feat_3"][0] - 1.4 < 0.00001)
    assert np.abs(phe.ds.df["feat_3"][1] - 5.9 < 0.00001)
    assert np.abs(phe.ds.df["feat_3"][2] - 0.34 < 0.00001)


def test_phe_getitem_setitem_delitem(dataset_iris):
    phe = Phenonaut(dataset_iris)
    phe["test_ds"] = dataset_iris.copy()
    assert len(phe.datasets) == 2
    assert len(phe[[0, 1]]) == 2
    assert phe["test_ds"].data.shape[1] == 4
    del phe["Iris"]
    assert len(phe.datasets) == 1
    assert phe.datasets[0].name == "test_ds"


def test_phe_save_and_load(dataset_iris):
    tmp_file = tempfile.NamedTemporaryFile(delete=True)
    try:
        phe = Phenonaut(dataset_iris)
        phe.save(tmp_file.name)
        phe_loaded = Phenonaut.load(tmp_file.name)
        assert len(phe_loaded.datasets) == 1
    finally:
        tmp_file.close()


def test_phe_clone(dataset_iris):
    phe = Phenonaut(dataset_iris)
    phe.clone_dataset("Iris", "Iris2")
    phe.clone_dataset("Iris", "Iris2", overwrite_existing=True)
    phe.clone_dataset("Iris", "Iris3", overwrite_existing=True)
    phe.clone_dataset("Iris", "Iris4", overwrite_existing=False)
    assert phe.keys() == ["Iris", "Iris2", "Iris3", "Iris4"]


def test_readin_transformations():
    df = pd.read_csv(
        StringIO(
            """"idx","Var1","Var2","value"
        
        "1","ENSG00000225972","A1_CPD1_PLATE1",4.0185
        "2","ENSG00000225630","A1_CPD1_PLATE1",1.1539
        "3","ENSG00000225972","A2_CPD2_PLATE1",10.6661
        "4","ENSG00000225630","A2_CPD2_PLATE1",1.6130
        "5","ENSG00000225972","A3_CPD3_PLATE1",0.1234
        "6","ENSG00000225630","A3_CPD3_PLATE1",9.8436
        "7","ENSG00000225972","A4_Pos_ctrl_PLATE1",0.1234
        "8","ENSG00000225630","A4_Pos_ctrl_PLATE1",9.8436"""
        )
    ).reset_index()
    phe = Phenonaut(
        df,
        "Test_df",
        metadata={
            "transforms": [
                ("replace_str", ("Var2", "Pos_", "Pos-")),
                ("split_column", ("Var2", "_", ["Well", "CpdID", "PlateID"])),
                ('pivot', ('Var1', 'value')),
                #'transpose',
                #(pivot_table(values = 'value', index=["WellName"], columns = 'Var1')
            ],
            "features_prefix":'ENSG',
        },
    )
    print(phe.df)



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from typing import Union, Optional
from pathlib import Path
import pkg_resources
import datetime

try:
    import pptx
except ImportError as ie:
        raise (f"Import Error - python-pptx functionality called, but it is not installed {ie}")


class PhenonautPPTX():
    """Phenonaut PPTX class
    
    Allows the creation of PPTX presentation files through insertion of images into slides.

    Instantiate the object using:
    
    Parameters
    ----------
    cover_title : str, optional
        Title to be displayed on the main title page of the presentation,
        by default "Phenonaut".
    cover_subtitle : str, optional
        Subtitle to be displayed on the main title page of the presentation,
        by default an empty string ("").
    template_file : Optional[Union[Path, str]], optional
        Template file that can be used to style the presentation. Phenonaut
        is supplied with a suitable template which is used as default when
        this argument is None. By default None.
    experiment_hash : Optional[str], optional
        Optionally supply a hash string which is displayed within the PPTX
        file. By default None.

    
    Then call add_image_slide to add slides, before finally calling save.
    """
    def __init__(self, cover_title:str="Phenonaut", cover_subtitle:str="",template_file:Optional[Union[Path, str]]=None, experiment_hash:Optional[str]=None):
        """Constructor of PhenonautPPTX object

        Parameters
        ----------
        cover_title : str, optional
            Title to be displayed on the main title page of the presentation,
            by default "Phenonaut".
        cover_subtitle : str, optional
            Subtitle to be displayed on the main title page of the presentation,
            by default an empty string ("").
        template_file : Optional[Union[Path, str]], optional
            Template file that can be used to style the presentation. Phenonaut
            is supplied with a suitable template which is used as default when
            this argument is None. By default None.
        experiment_hash : Optional[str], optional
            Optionally supply a hash string which is displayed within the PPTX
            file. By default None.
        """

        self.top_left_content_loc=(pptx.util.Cm(2.33), pptx.util.Cm(3.31))
        
        if template_file is None:
            template_file = pkg_resources.resource_filename('phenonaut', 'output/ph0_pptx_template.pptx')
        if isinstance(template_file, str):
            template_file=Path(template_file)

        self.pres=pptx.Presentation(template_file)
        slide=self.pres.slides.add_slide(self.pres.slide_layouts[0])
        slide.shapes.title.text = cover_title
        slide.placeholders[1].text=cover_subtitle
        slide.placeholders[13].text = f"Generated: {datetime.datetime.now().strftime('%d/%m/%Y, %H:%M:%S')}"
        slide.placeholders[14].text = experiment_hash if experiment_hash is not None else "Hash: NA"

    def add_image_slide(self, title:str, img_file:Union[Path, str],width:Optional[Union[int, float]]=None, height:Optional[Union[int, float]]=None):
        """Add an image slide to an existing presentation

        Parameters
        ----------
        title : str
            Slide title
        img_file : Union[Path, str]
            Image file to insert into slide
        width : Optional[Union[int, float]], optional
            Width in cm to scale image. No scaling is done if argument is None,
            by default None.
        height : Optional[Union[int, float]], optional
            Height in cm to scale image. No scaling is done if argument is None,
            by default None.
        """        
        if isinstance(img_file, str):
            img_file=Path(img_file)
        if isinstance(width, (int, float)):
            width=pptx.util.Cm(width)
        if isinstance(height, (int, float)):
            height=pptx.util.Cm(height)
        slide=self.pres.slides.add_slide(self.pres.slide_layouts[1])
        slide.placeholders[0].text=title
        slide.shapes.add_picture(str(img_file), *self.top_left_content_loc, width=width, height=height)
        

    def save(self, output_pptx_file:Union[Path, str]):
        """Save the presentation

        Parameters
        ----------
        output_pptx_file : Union[Path, str]
            Output file path
        """        
        self.output_pptx_file = output_pptx_file
        if isinstance(self.output_pptx_file, str):
            self.output_pptx_file=Path(self.output_pptx_file)
        self.pres.save(self.output_pptx_file)




# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from matplotlib import pyplot as plt
from pathlib import Path
from typing import Union, Optional
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

def write_boxplot_to_file(
    df:pd.DataFrame,
    x_label_in_df:str,
    y_label_in_df:str,
    output_file: Union[str, Path],
    title="Boxplot",
    x_label:Optional[str]=None,
    y_label:Optional[str]=None,
    figsize=(10, 6.18),
    ):
    """Write a boxplot to a file

    Boxplot generation from pd.DataFrames using the Seaborn library

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame
    x_label_in_df : str
        Label for the x-axis which appears in the dataframe.
    y_label_in_df : str
        Label for the y-axis which appears in the dataframe.
    output_file : Union[str, Path]
        Output (PNG) filename.
    title : str, optional
        Title to be displayed on the plot, by default "Boxplot"
    x_label : Optional[str], optional
        Optionally, override the use of the x-label present in the dataframe
        and display a different one on the axis. If None, the the value of
        x_label_in_df is used. By default None.
    y_label : Optional[str], optional
        Optionally, override the use of the y-label present in the dataframe
        and display a different one on the axis. If None, the the value of
        y_label_in_df is used. By default None.
    figsize : tuple, optional
        Optional tuple of image height-width (in inches - as dictated by
        matplotlib/seaborn), by default (10, 6.18).
    """    
    sns.set(rc={'figure.figsize':figsize})
    tidy_df = df.replace(' ',r'\n', regex=True) 

    if isinstance(output_file, str):
        output_file=Path(output_file)
    if not output_file.parent.exists():
        output_file.parent.mkdir(parents=True)
    ax=sns.boxplot(x=x_label_in_df, y=y_label_in_df, data=tidy_df)
    if x_label is not None:
        ax.set(xlabel=x_label)
    if y_label is not None:
        ax.set(ylabel=y_label)
    
    ax.set_title(title)
    fig = ax.get_figure()
    plt.tight_layout()
    fig.savefig(output_file)
    plt.close()



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from typing import List, Tuple, Union, Optional

from pandas.errors import DataError

from phenonaut.data import Dataset, dataset
import pandas as pd

from scipy.spatial.distance import pdist, cdist, cityblock
from scipy.linalg import inv
import numpy as np
from scipy.spatial import distance
from sklearn.covariance import EmpiricalCovariance, MinCovDet


def treatment_spread_euclidean(
    data: Dataset,
    perturbation_column: Optional[str] = None,
    perturbations: Optional[List[str]] = None,
) -> dict[str:float]:
    """Calculate the euclidean spreat of perturbation repeats

    Returns
    -------
    dict[str:float]
        Dictionary with perturbations as keys and euclidean distances as values.

    Raises
    ------
    DataError
        Perturbation column was not set for the given Dataset and was not
        supplied via the perturbation_column argument to
        treatment_spread_euclidean.
    """    
    if perturbation_column is None and data.perturbation_column is None:
        raise DataError("Dataset does not have a perturbation column set and none was supplied to treatment_spread_euclidean")
    if perturbation_column is None:
        perturbation_column=data.perturbation_column
    perturbations = data.get_unique_perturbations() if perturbations is None else perturbations
    
    return {
        t: np.sum(
            pdist(dataset.df.query(f"{data.perturbation_column} == '{t}'")[data.features]),
            axis=None,
        )
        for t in perturbations
    }


def mahalanobis(
    point: Union[List[float], np.ndarray, pd.DataFrame],
    cloud: Union[List[List[float]], np.ndarray, pd.DataFrame],
    pvals:bool=False,
    covariance: Union[np.ndarray, EmpiricalCovariance, MinCovDet, None] = EmpiricalCovariance(),
):
    """Measure the Mahalanobis distance between a point and a cloud

    The Mahalanobis distance calculation is particularly sensitive to outliers,
    which results in large changes in covariance matrix calculation. For this
    reason, robust covarience estimators may be suppplied to the method. Whilst
    a common recomendation is to calculate the sqrt of the Mahalanobis distance,
    this is an approximation to euclidean space, only correct when the covariance
    matrix is an identity matrix. As Phenonaut is concerned on operating on high
    dimensional space which very likely has covariences present, the returned
    distance is not square rooted, returning D2 as noted.
    
    https://imaging.mrc-cbu.cam.ac.uk/statswiki/FAQ/euclid

    Optionally, the p-value for the point or points belonging to the cloud can
    be returned.

    Parameters
    ----------
    point : Union[List[float],np.ndarray, pd.DataFrame]
        Multidimensional point, can be a simple list of features, or a 2d M*N
        list where M are multiple points to be individually measured of N
        features and returning an array of measurmenents for each point.
    cloud : Union[List[List[float]], np.ndarray, pd.DataFrame]
        2D M*N array-like set of M points, with N features from which the
        underlying target distribution will be measured.
    pvals : bool
        If True, then p-value for the point (or points) belonging to cloud is returned.
        This is calculated using Chi2 and degrees of freedom calculated as
        N-1, where N is the number of features/dimensions. If point was one
        dimensional, then a single floating point value is returned. If it is 2D
        (MxN matrix), then an array of length M is returned. By default, False.
    covariance : Optional[Union[np.ndarray, EmpiricalCovariance]], optional
        If none, then the covariance matrix is calculated using scikit's
        EmpiricalCovariance. This is fairly robust to outliers, and much more robust
        than the standard approach to calculating a covariance matrix (using numpy's
        np.cov method. Robust estimators may be used by passing in an instantiated
        object which a subclass of EmpiracalCovariance (like
        sklearn.covariance.MinCovDet or EmpiricalCovariance itself). If None, then
        numpy's cov method is used (sensitive to outliers). By default EmpiracalCovariance().

    Returns
    -------
    [np.ndarray, float]
        Array (if point is 2D) of distances. If 1D, then single float value is
        returned indicating the Mahalanobis distance of the point to the
        cloud.

    """
    point=np.array(point)
    cloud=np.array(cloud)
    if point.ndim==1:
        point=point.reshape(1,-1)
    
    if isinstance(covariance,EmpiricalCovariance):
        covariance.fit(cloud)
        mahl_d=covariance.mahalanobis(point)
    else:
        if covariance is None:
            cov_mat = np.cov(cloud.T)
        elif isinstance(covariance, np.ndarray):
            cov_mat=covariance
        else:
            raise ValueError(f"cov argument should be one of np.ndarray, None, or a subclass of sklearn.covariance.EmpiricalCovariance, it was {type(covariance)}")
        difference = point - np.mean(cloud, axis=0)
        mahl_d=np.diag(np.dot(np.dot(difference, np.linalg.inv(cov_mat)), difference.T))
    if len(mahl_d)==1:
        mahl_d=mahl_d[0]
    if not pvals:
        return mahl_d
    else:
        from scipy.stats import chi2
        return 1-chi2.cdf(mahl_d,cloud.shape[1]-1)


def euclidean(
    point1: Union[List[float], np.ndarray, pd.DataFrame],
    point2: Union[List[List[float]], np.ndarray, pd.DataFrame],
):
    """Measure the euclidean distance between 2 points

    Parameters
    ----------
    point1 : Union[List[float],np.ndarray, pd.DataFrame]
        Multidimensional point, can be a simple list of [x,y,z], or a 2d M*N
        list where M are multiple points to be individually measured of N
        features and returning an array of measurmenents for each point.
    point2 : Union[List[List[float]], np.ndarray, pd.DataFrame]
        List of N features, must have 1 dimension.

    Returns
    -------
    [np.ndarray, float]
        Array (if point is 2D) of distances. If 1D, then single float value is
        returned indicating the euclidean distance between point1 and point2.

    Raises
    ------
    DataError
        Point1 can be 2D, but not 3D, and point2 must be 1D
    """
    point1 = np.array(point1)
    point2 = np.array(point2)
    if point1.ndim == 1:
        point1 = point1.reshape(1, -1)
    if point1.ndim > 2:
        raise DataError(
            f"point1 can be a 2D M*N array where M is unique points, and N are features. point1 shape was {point1.shape}."
        )
    if point2.ndim > 1:
        raise DataError(f"point2 must be 1D, it was {point2.ndim}.")
    results = np.empty(point1.shape[0])
    for i, p in enumerate(point1):
        results[i] = np.linalg.norm(point2 - p)
    if len(results) == 1:
        return np.sqrt(results[0])
    else:
        return results


def manhattan(
    point1: Union[List[float], np.ndarray, pd.DataFrame],
    point2: Union[List[List[float]], np.ndarray, pd.DataFrame],
):
    """Measure the Manhattan distance between 2 points

    Parameters
    ----------
    point1 : Union[List[float],np.ndarray, pd.DataFrame]
        Multidimensional point, can be a simple list of [x,y,z], or a 2d M*N
        list where M are multiple points to be individually measured of N
        features and returning an array of measurmenents for each point.
    point2 : Union[List[List[float]], np.ndarray, pd.DataFrame]
        List of N features, must have 1 dimension.

    Returns
    -------
    [np.ndarray, float]
        Array (if point is 2D) of distances. If 1D, then single float value is
        returned indicating the euclidean distance between point1 and point2.

    Raises
    ------
    DataError
        Point1 can be 2D, but not 3D, and point2 must be 1D
    """
    point1 = np.array(point1)
    point2 = np.array(point2)
    if point1.ndim == 1:
        point1 = point1.reshape(1, -1)
    if point1.ndim > 2:
        raise DataError(
            f"point1 can be a 2D M*N array where M is unique points, and N are features. point1 shape was {point1.shape}."
        )
    if point2.ndim > 1:
        raise DataError(f"point2 must be 1D, it was {point2.ndim}.")
    results = np.empty(point1.shape[0])
    for i, p in enumerate(point1):
        results[i] = cityblock(p, point2)
    if len(results) == 1:
        return np.sqrt(results[0])
    else:
        return results



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from phenonaut import Phenonaut
from phenonaut.data import Dataset
from phenonaut.transforms.preparative import VIF, RemoveHighlyCorrelated, RemoveHighestCorrelatedThenVIF
import pandas as pd
import pytest
from pathlib import Path

@pytest.fixture
def sklearn_regression_df():
    """Return sklearn regression dataframe generated by sklearn.datasets.make_regression
    
    sklearn generated regression datasets, 100 rows, 100 features (named feat_n where n
    is 1-100), and one target column which is a regression target.
    
    Generated using the following python code:
        from sklearn.datasets import make_regression
        X,y=make_regression(random_state=42)
        df = pd.DataFrame(np.hstack([X,y.reshape(-1,1)]), columns=[f"feat_{i+1}" for i in range(X.shape[1])]+["target"])
        df.to_csv("test/generated_regression_dataset.csv")

    """
    return pd.read_csv(Path("test")/"generated_regression_dataset.csv", index_col=0)


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_VIF(sklearn_regression_df):
    phe=Phenonaut(sklearn_regression_df, "PheSmallDS", metadata={'features_prefix':'feat_'})
    vif=VIF()
    vif.filter(phe.ds, vif_cutoff=2, min_features=50)
    assert len(phe.ds.features)==52
    phe=Phenonaut(sklearn_regression_df, "PheSmallDS", metadata={'features_prefix':'feat_'})
    vif=VIF()
    vif(phe.ds, vif_cutoff=2, min_features=60)
    assert len(phe.ds.features)==79

def test_RemoveHighlyCorrelated(sklearn_regression_df):
    phe=Phenonaut(sklearn_regression_df, "PheSmallDS", metadata={'features_prefix':'feat_'})
    rhc=RemoveHighlyCorrelated()
    rhc(phe.ds, threshold=0.3)
    assert len(phe.ds.features)==95
    phe=Phenonaut(sklearn_regression_df, "PheSmallDS", metadata={'features_prefix':'feat_'})
    rhc=RemoveHighlyCorrelated()
    rhc(phe.ds, threshold=0.4, min_features=97)
    assert len(phe.ds.features)==97
    phe=Phenonaut(sklearn_regression_df, "PheSmallDS", metadata={'features_prefix':'feat_'})
    rhc=RemoveHighlyCorrelated()
    rhc(phe.ds, threshold=None, min_features=5)
    assert len(phe.ds.features)==5

def test_RemoveHighestCorrelatedThenVIF(sklearn_regression_df):
    phe=Phenonaut(sklearn_regression_df, "PheSmallDS", metadata={'features_prefix':'feat_'})
    rhcvif=RemoveHighestCorrelatedThenVIF()
    rhcvif(phe.ds, n_before_vif=95, vif_cutoff=5.0)
    assert len(phe.ds.features)==78


# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from copy import deepcopy
from sklearn.model_selection import KFold
from phenonaut.predict.predictor_dataclasses import PhenonautPredictionMetric, PhenonautPredictor
import numpy as np
import optuna
from typing import Optional, Union
from pickle import dumps as pickle_dumps
from pickle import loads as pickle_loads
from timeit import default_timer
import base64
import gzip
from pathlib import Path
import phenonaut

from phenonaut.predict.predict_utils import PredictionType


def predictor_to_str(ob: PhenonautPredictor) -> str:
    """Encode a PhenonautPredictor (or any serialisable object) to str

    First serialise the object using pickle.dumps, compress using gzip and
    return the base64 string representation.  Works with any serialisable
    object.

    Parameters
    ----------
    ob : PhenonautPredictor
        Serializable object to convert.

    Returns
    -------
    str
        Base64 gzip serialised object as a utf-8 string.
    """
    return base64.b64encode(gzip.compress(pickle_dumps(ob))).decode("utf-8")


def predictor_from_str(ob_str: str) -> PhenonautPredictor:
    """Convert a base64 gziped object string to PhenonautPredictor

    Convert a str to object after applying gzip decompression and then
    pickle.loads.

    Parameters
    ----------
    ob_str : str
        String representation of base64 encoded gzipped object

    Returns
    -------
    PhenonautPredictor
        PhenonautPredictor object
    """
    return pickle_loads(gzip.decompress(base64.b64decode(ob_str)))


def run_optuna_opt(
    X: Union[list[np.ndarray], np.ndarray],
    X_test: Union[list[np.ndarray], np.ndarray],
    y: Union[list[np.ndarray], np.ndarray],
    y_test: Union[list[np.ndarray], np.ndarray],
    prediction_type: PredictionType,
    predictor: PhenonautPredictor,
    metric: PhenonautPredictionMetric,
    n_optuna_trials: int,
    phe_name: str,
    dataset_combination: list[str],
    optuna_db_path: Union[Path, str],
    n_splits: int = 5,
    seed: Optional[int] = None,
    optuna_db_protocol: str = "sqlite:///",
    target_dataset_name: Optional[str] = None,
):
    """Run Optuna-led hyperparameter optimisation on predictor and data

    Parameters
    ----------
    X : Union[list[np.ndarray], np.ndarray]
        Training data
    X_test : Union[list[np.ndarray], np.ndarray]
        Test data
    y : Union[list[np.ndarray], np.ndarray]
        Training target
    y_test : Union[list[np.ndarray], np.ndarray]
        Test target data
    prediction_type : PredictionType
        Prediction type specifying classification, regression or view.
    predictor : PhenonautPredictor
        The predictor with fit, predict functions packed into PhenonautPredictor
        dataclass which supports the predictor class with specification of
        hyperparameters etc.
    metric : PhenonautPredictionMetric
        The scoring metric to be used to assess performance.
    n_optuna_trials : int
        Number of optuna trials to optimise across.
    phe_name : str
        Name of the Phenonaut object from which the data comes
    dataset_combination : list[str]
        Combination views of datasets to be used
    optuna_db_path : Union[Path, str]
        Output file path for Optuna sqlite3 database file
    n_splits : int, optional
        Number of splits to be used in cross fold validation, by default 5
    seed : Optional[int], optional
        Seed for use by random number generator, allow deterministing repeats.
        If None, then do no pre-seed the generator. By default None
    optuna_db_protocol : _type_, optional
        Protocol for optuna to use to access storage. By default "sqlite:///"
    target_dataset_name : Optional[str], optional
        If predicting a view, then the target Dataset may be given a name,
        by default None.
    """    
    
    # First, we make a deepcopy of predictor, as the type is mutable, and we will be changing it in
    # this function, adding a member variable to obtain the trained predictor back from the
    # augmented optuna objective function.
    predictor=deepcopy(predictor)
    # If predictor has max_optuna_trials set lower than max_optuna trials supplied by the function arg,
    # then take the lower. If not set, then we can use all trials denoted by n_optuna_trials argument.
    if predictor.max_optuna_trials is None:
        n_optuna_trials_for_predictor = n_optuna_trials
    else:
        n_optuna_trials_for_predictor = min(predictor.max_optuna_trials, n_optuna_trials)

    for fold_index, (train_indexes, validation_indexes) in enumerate(
        KFold(n_splits=n_splits, shuffle=True, random_state=seed).split(X[0] if isinstance(X, list) else X)
    ):
        if isinstance(X, list):
            X_train = [X[i][train_indexes] for i in range(len(X))]
            X_val = [X[i][validation_indexes] for i in range(len(X))]
        else:
            X_train = X[train_indexes]
            X_val = X[validation_indexes]
        y_train = y[train_indexes]
        y_val = y[validation_indexes]

        # Name the study- must be unique for predictor/dataset/target predictions
        if prediction_type == PredictionType.view and target_dataset_name is not None:
            optuna_study_name = (
                f"{phe_name};{prediction_type.name};{predictor.name};{dataset_combination};{target_dataset_name};fold{fold_index+1}"
            )
        else:
            optuna_study_name = f"{phe_name};{prediction_type.name};{predictor.name};{dataset_combination};fold{fold_index+1}"

        # Load study
        optuna_study = optuna.create_study(
            study_name=optuna_study_name,
            storage=f"{optuna_db_protocol}{optuna_db_path}",
            load_if_exists=True,
            direction="minimize" if metric.lower_is_better else "maximize",
        )
        # Some predictors like GaussianNB do not take any hyperparameters, others are easily explored with
        # some categorical parameters. These therefore require 1 run or very few to explore the entire search
        # space. We therefore see if max_optuna_trials for the predictor has been reached in this study before
        # if so, then there is no reason to test those parameters again.
        if predictor.max_optuna_trials is not None and len(optuna_study.trials) >= predictor.max_optuna_trials:
            print(f"max_optuna_trials ({predictor.max_optuna_trials}) already reached for study, continuing.")
            return

        # It is a new study, set the user attributes for ease of analysis
        if len(optuna_study.trials) == 0:
            optuna_study.set_user_attr("phe_obj_name", phe_name)
            optuna_study.set_user_attr("predictor_name", predictor.name)
            optuna_study.set_user_attr("dataset", ":".join(dataset_combination))
            optuna_study.set_user_attr("prediction_type", prediction_type.name)
            if prediction_type == PredictionType.view and target_dataset_name is not None:
                optuna_study.set_user_attr("predicting_view", target_dataset_name)
            optuna_study.set_user_attr("metric_name", metric.name)
            optuna_study.set_user_attr("lower_is_better", metric.lower_is_better)
            optuna_study.set_user_attr("merge_folds", False)
            optuna_study.set_user_attr("fold_index", fold_index+1)
        start_time=default_timer()
        optuna_study.optimize(
            lambda trial: _optuna_augmented_objective(
                trial,
                optuna_study,
                X_train,
                X_val,
                y_train,
                y_val,
                predictor,
                metric,
            ),
            n_trials=n_optuna_trials_for_predictor,
        )
        optuna_study.set_user_attr("time_taken", default_timer()-start_time)
        test_score = metric(y_test, predictor._best_predictor_instance.predict(X_test))
        optuna_study.set_user_attr("test_score", test_score)


def _optuna_augmented_objective(
    trial: optuna.Trial,
    optuna_study: optuna.Study,
    X_train: Union[list[np.ndarray], np.ndarray],
    X_val: Union[list[np.ndarray], np.ndarray],
    y_train: np.ndarray,
    y_val: np.ndarray,
    predictor: PhenonautPredictor,
    metric: PhenonautPredictionMetric,
):
    """Flexible Optuna objective

    Parameters
    ----------
    trial : optuna.Trial
        The Optua trial passed by Optuna
    optuna_study : optuna.Study
        Study name
    X_train : Union[list[np.ndarray], np.ndarray]
        Training data
    X_val : Union[list[np.ndarray], np.ndarray]
        Validation data
    y_train : np.ndarray
        Training target data
    y_val : np.ndarray
        Validation target data
    predictor : PhenonautPredictor
        Predictor dataclass holding predictor and possible hyperparameters.
    metric : PhenonautPredictionMetric
        The metric to be used to score predictions.

    Returns
    -------
    float
        Score of trained model.
    """
    optuna_dict = {}
    if predictor.optuna is None:
        optuna_dict = {}
    else:
        optuna_dict = {
            hyperparameter.name: getattr(trial, hyperparameter.optuna_func)(
                hyperparameter.name, *hyperparameter.parameters
            )
            for hyperparameter in predictor.optuna
        }

    if predictor.conditional_hyperparameter_generator is not None:
        optuna_dict.update(
            predictor.conditional_hyperparameter_generator(
                predictor.conditional_hyperparameter_generator_constructor_keyword,
                trial,
                optuna_dict,
                X_train.shape[1],
                y_train.shape[1] if y_train.ndim == 2 else 1,
            )
        )
    if isinstance(predictor.predictor, tuple):
        predictor_instance = predictor.predictor[0](predictor.predictor[1](**optuna_dict, **predictor.constructor_kwargs))
    else:
        predictor_instance = predictor.predictor(**optuna_dict, **predictor.constructor_kwargs)
    predictor_instance.fit(X_train, y_train)
    score=metric(y_val, predictor_instance.predict(X_val))
    
    # If not yet set (because its a new study with 0 trials and was initialised to None, then set it to the current score)
    if "best_trained_model_score" not in optuna_study.user_attrs:
        optuna_study.set_user_attr("best_trained_model_score", score)
    if not hasattr(predictor, "_best_predictor_instance"):
        predictor._best_predictor_instance=predictor_instance

    cur_best_score = optuna_study.user_attrs["best_trained_model_score"]

    if (metric.lower_is_better and score <= cur_best_score) or (
        not metric.lower_is_better and score >= cur_best_score
    ):
        optuna_study.set_user_attr("best_trained_model_score", score)
        predictor._best_predictor_instance=predictor_instance

    return score



























def run_optuna_opt_merge_folds(
    X: Union[list[np.ndarray], np.ndarray],
    X_test: Union[list[np.ndarray], np.ndarray],
    y: np.ndarray,
    y_test: np.ndarray,
    prediction_type: PredictionType,
    predictor: PhenonautPredictor,
    metric: PhenonautPredictionMetric,
    n_optuna_trials: int,
    phe_name: str,
    dataset_combination: list[str],
    optuna_db_path: Union[Path, str],
    n_splits: int = 5,
    seed: Optional[int] = None,
    optuna_db_protocol: str = "sqlite:///",
    target_dataset_name: Optional[str] = None,
):
    # First, we make a deepcopy of predictor, as the type is mutable, and we will be changing it in
    # this function, adding a member variable to obtain the trained predictor back from the
    # augmented optuna objective function.
    predictor=deepcopy(predictor)
    # If predictor has max_optuna_trials set lower than max_optuna trials supplied by the function arg,
    # then take the lower. If not set, then we can use all trials denoted by n_optuna_trials argument.
    if predictor.max_optuna_trials is None:
        n_optuna_trials_for_predictor = n_optuna_trials
    else:
        n_optuna_trials_for_predictor = min(predictor.max_optuna_trials, n_optuna_trials)

    # Name the study- must be unique for predictor/dataset/target predictions
    if prediction_type == PredictionType.view and target_dataset_name is not None:
        optuna_study_name = f"{phe_name};{prediction_type.name};{predictor.name};{dataset_combination};{target_dataset_name};merge_folds"
    else:
        optuna_study_name = f"{phe_name};{prediction_type.name};{predictor.name};{dataset_combination};merge_folds"

    # Load study
    optuna_study = optuna.create_study(
        study_name=optuna_study_name,
        storage=f"{optuna_db_protocol}{optuna_db_path}",
        load_if_exists=True,
        direction="minimize" if metric.lower_is_better else "maximize",
    )
    # Some predictors like GaussianNB do not take any hyperparameters, others are easily explored with
    # some categorical parameters. These therefore require 1 run or very few to explore the entire search
    # space. We therefore see if max_optuna_trials for the predictor has been reached in this study before
    # if so, then there is no reason to test those parameters again.
    if predictor.max_optuna_trials is not None and len(optuna_study.trials) >= predictor.max_optuna_trials:
        print(f"max_optuna_trials ({predictor.max_optuna_trials}) already reached for study, continuing.")
        return

    # It is a new study, set the user attributes for ease of analysis
    if len(optuna_study.trials) == 0:
        optuna_study.set_user_attr("phe_obj_name", phe_name)
        optuna_study.set_user_attr("predictor_name", predictor.name)
        optuna_study.set_user_attr("dataset", ":".join(dataset_combination))
        optuna_study.set_user_attr("prediction_type", prediction_type.name)
        if prediction_type == PredictionType.view and target_dataset_name is not None:
            optuna_study.set_user_attr("predicting_view", target_dataset_name)
        optuna_study.set_user_attr("metric_name", metric.name)
        optuna_study.set_user_attr("lower_is_better", metric.lower_is_better)
        optuna_study.set_user_attr("merge_folds", True)
    start_time=default_timer()
    optuna_study.optimize(
        lambda trial: _optuna_augmented_objective_merge_folds(
            trial, optuna_study, X, y, predictor, metric, n_splits, seed=seed
        ),
        n_trials=n_optuna_trials_for_predictor,
    )
    optuna_study.set_user_attr("time_taken", default_timer()-start_time)
    test_score = metric(y_test, predictor._best_predictor_instance.predict(X_test))
    optuna_study.set_user_attr("test_score", test_score)


def _optuna_augmented_objective_merge_folds(
    trial: optuna.Trial,
    optuna_study: optuna.Study,
    X: Union[list[np.ndarray], np.ndarray],
    y: np.ndarray,
    predictor: PhenonautPredictor,
    metric: PhenonautPredictionMetric,
    n_splits: int = 5,
    seed=None,
):
    scores = []
    optuna_dict = {}
    if predictor.optuna is None:
        optuna_dict = {}
    else:
        optuna_dict = {
            hyperparameter.name: getattr(trial, hyperparameter.optuna_func)(
                hyperparameter.name, *hyperparameter.parameters
            )
            for hyperparameter in predictor.optuna
        }

    if predictor.conditional_hyperparameter_generator is not None:
        optuna_dict.update(
            predictor.conditional_hyperparameter_generator(
                predictor.conditional_hyperparameter_generator_constructor_keyword,
                trial,
                optuna_dict,
                X.shape[1],
                y.shape[1] if y.ndim == 2 else 1,
            )
        )

    if isinstance(predictor.predictor, tuple):
        predictor_instance = predictor.predictor[0](predictor.predictor[1](**optuna_dict))
    else:
        predictor_instance = predictor.predictor(**optuna_dict)

    # multi view
    if isinstance(X, list):
        for train_indexes, validation_indexes in KFold(n_splits=n_splits, shuffle=True, random_state=seed).split(X[0]):
            mv_X_train = [X[i][train_indexes] for i in range(len(X))]
            mv_X_test = [X[i][validation_indexes] for i in range(len(X))]
            predictor_instance.fit(mv_X_train, y[train_indexes])
            y_pred = predictor_instance.predict(mv_X_test)
            scores.append(metric(y[validation_indexes], y_pred))
    else:
        for train_indexes, validation_indexes in KFold(n_splits=n_splits, shuffle=True, random_state=seed).split(X):
            predictor_instance.fit(X[train_indexes], y[train_indexes])
            y_pred = predictor_instance.predict(X[validation_indexes])
            scores.append(metric(y[validation_indexes], y_pred))

    # If not yet set (because its a new study with 0 trials and was initialised to None, then set it to the current score)
    if "best_trained_model_score" not in optuna_study.user_attrs:
        optuna_study.set_user_attr("best_trained_model_score", np.mean(scores))
    if not hasattr(predictor, "_best_predictor_instance"):
        predictor._best_predictor_instance=predictor_instance
    cur_best_score = optuna_study.user_attrs["best_trained_model_score"]

    if (metric.lower_is_better and np.mean(scores) <= cur_best_score) or (
        not metric.lower_is_better and np.mean(scores) >= cur_best_score
    ):
        optuna_study.set_user_attr("best_trained_model_score", np.mean(scores))
        predictor._best_predictor_instance=predictor_instance
    trial.set_user_attr("KFoldScores", scores)
    trial.set_user_attr("Mean_KFold", np.mean(scores))
    trial.set_user_attr("Std_KFold", np.std(scores))

    return np.mean(scores)
    


# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from abc import ABC, abstractclassmethod, abstractmethod
from os import path
from pathlib import Path
from time import sleep
from typing import Callable, List, Tuple, Union, Optional
import urllib.request
from urllib.error import URLError
from progressbar import progressbar
import shutil
import gzip


class PackagedDataset(ABC):
    """PackagedDataset base class for all downloaded Datasets

    Inherited by Phenonaut classes which supply public datasets in the same
    way that pytorch allows easy access to MNIST and FashionMNIST etc,
    Phenonaut offers classes which download and preprocess datasets like
    TCGA (The Cancer Genome Atlas and the Connectivity Map), which may
    include many different 'views' or omics-based measurements of the
    underlying cells.

    Inheriting from this class allows easy access to commonly used functions
    for checking datasets exist in directories and downloading them if not,
    in addition to more small helpful functions. Inheriting grants the
    following:

    * Getters and setters for root and raw_data_dir, properly handling the expected location of dataset files, listing available Phenonaut Dataset objects via the .keys() or ds_keys() methods, listing supporting dataframes via the df_keys() methods.
    * download and batch_download functions which simplify the download of remote public datasets.
    * processed_dataset_exists and raw_dataset_exists, which check for the presence of the processed dataset and the raw dataset, respectively.

    Inheriting classes should do the following:

    * Call super().__init__() on initialisation
    * Check if they find a saved/processed version of the packaged dataset. The CMAP and TCGA classes which inherit from this PackagedDataset process and save the datasets in an h5 file. This is optional, and any store may be used.
    * If it does not exist, download the data storing it in .raw_data_dir, process the data and store in a convenient format.
    * Register the available Phenonaut Datasets associated with this PackagedDataset.  By convention, the default/main dataset should be named 'ds'. Registration is completed by calling self.register_ds_key('ds_name'). Available Phenonaut Datasets are available by calling self.keys(), or self.ds_keys(). Phenonaut datasets may be accessed by calling get_ds, or with the ['ds_name'] notation on the PackagedDataset instance.
    * Register the available supporting dataframes associated with this PackagedDataset. Supporting dataframe names can be listed by calling df_keys() and accessed by calling get_df('df_name').
    * Classes should provide their own get_df and get_ds methods. This is enforced by this base class specifying required methods to be present through inheritance of the AbstractBaseClass.

    Parameters
    ----------
    root : Union[Path, str]
        Root directory for the dataset. The root directory should contain
        processed files, usable by Phenonaut, this means that the data has
        been downloaded and usually transformed in some manner prior to
        being put here. By convention, processed files will be put into
        this directory, but there will exist a subdirectory called
        "raw_data", within which downloaded files (possibly compressed)
        will be placed prior to preprocessing.
    raw_data_dir : Optional[Union[Path, str]], optional
        Directory in which the raw, downloaded files should be saved, also
        the location of intermediate files generated in the processing step.
        By convention, this directory lies within the root directory and has
        the default name "raw_data". It can be an absolute path in a
        different directory or filesystem by setting
        raw_data_dir_relative_to_root argument to False.
        By default Path("raw_data").
    raw_data_dir_relative_to_root : bool, optional
        As described for the raw_data_dir argument, by the "raw_data"
        directory is usually within the root directory for the dataset.
        If this argument is True, then the full path of the raw_data_dir is
        generated with root as the parent directory.  If this value is
        False, then it is taken to be an absolute path, possibly existing
        on another filesystem. By default True.
        """

    def __init__(
        self,
        root: Union[Path, str],
        raw_data_dir: Optional[Union[Path, str]] = Path("raw_data"),
        raw_data_dir_relative_to_root: bool = True,
        download: bool = True,
    ):
        """PackagedDataset base class for all downloaded Datasets

        Inherited by Phenonaut classes which supply public datasets in the same
        way that pytorch allows easy access to MNIST and FashionMNIST etc,
        Phenonaut offers classes which download and preprocess datasets like
        TCGA (The Cancer Genome Atlas and the Connectivity Map), which may
        include many different 'views' or omics-based measurements of the
        underlying cells.

        Inheriting from this class allows easy access to commonly used functions
        for checking datasets exist in directories and downloading them if not,
        in addition to more small helpful functions. Inheriting grants the
        following:

        * Getters and setters for root and raw_data_dir, properly handling the expected location of dataset files, listing available Phenonaut Dataset objects via the .keys() or ds_keys() methods, listing supporting dataframes via the df_keys() methods.
        * download and batch_download functions which simplify the download of remote public datasets.
        * processed_dataset_exists and raw_dataset_exists, which check for the presence of the processed dataset and the raw dataset, respectively.

        Inheriting classes should do the following:

        * Call super().__init__() on initialisation
        * Check if they find a saved/processed version of the packaged dataset. The CMAP and TCGA classes which inherit from this PackagedDataset process and save the datasets in an h5 file. This is optional, and any store may be used.
        * If it does not exist, download the data storing it in .raw_data_dir, process the data and store in a convenient format.
        * Register the available Phenonaut Datasets associated with this PackagedDataset.  By convention, the default/main dataset should be named 'ds'. Registration is completed by calling self.register_ds_key('ds_name'). Available Phenonaut Datasets are available by calling self.keys(), or self.ds_keys(). Phenonaut datasets may be accessed by calling get_ds, or with the ['ds_name'] notation on the PackagedDataset instance.
        * Register the available supporting dataframes associated with this PackagedDataset. Supporting dataframe names can be listed by calling df_keys() and accessed by calling get_df('df_name').
        * Classes should provide their own get_df and get_ds methods. This is enforced by this base class specifying required methods to be present through inheritance of the AbstractBaseClass.

        Parameters
        ----------
        root : Union[Path, str]
            Root directory for the dataset. The root directory should contain
            processed files, usable by Phenonaut, this means that the data has
            been downloaded and usually transformed in some manner prior to
            being put here. By convention, processed files will be put into
            this directory, but there will exist a subdirectory called
            "raw_data", within which downloaded files (possibly compressed)
            will be placed prior to preprocessing.
        raw_data_dir : Optional[Union[Path, str]], optional
            Directory in which the raw, downloaded files should be saved, also
            the location of intermediate files generated in the processing step.
            By convention, this directory lies within the root directory and has
            the default name "raw_data". It can be an absolute path in a
            different directory or filesystem by setting
            raw_data_dir_relative_to_root argument to False.
            By default Path("raw_data").
        raw_data_dir_relative_to_root : bool, optional
            As described for the raw_data_dir argument, by the "raw_data"
            directory is usually within the root directory for the dataset.
            If this argument is True, then the full path of the raw_data_dir is
            generated with root as the parent directory.  If this value is
            False, then it is taken to be an absolute path, possibly existing
            on another filesystem. By default True.
        """
        self.root = root
        self._raw_data_dir_relative_to_root = raw_data_dir_relative_to_root
        self.raw_data_dir = raw_data_dir
        self._keys_to_accessible_dfs=[]
        self._keys_to_accessible_dss=[]

    @property
    def root(self) -> Path:
        """Get the root directory of the dataset

        Returns
        -------
        Path
            Root directory within which the processed dataset and raw dataset
            data can be found.
        """
        return self._root
    
    @root.setter
    def root(self, val: Union[Path, str]):
        """Set the root directory of the dataset

        Parameters
        -------
        val :  Union[Path, str]
            Root directory within which the processed dataset and raw dataset
            data can be found.
        """
        if isinstance(val, str):
            val = Path(val)
        if not isinstance(val, Path):
            raise TypeError(f"root should be of type Path or str, but was {type(val)}")
        self._root = val

    @property
    def raw_data_dir(self) -> Path:
        """Getthe raw unprocessed data directory of the dataset

        Returns
        -------
        Path
            Directory of the raw, unprocessed dataset.
        """
        return self._raw_data_dir

    @raw_data_dir.setter
    def raw_data_dir(self, val:Union[Path, str]):
        """Set the raw unprocessed data directory of the dataset

        Parameters
        -------
        Path : Union[Path, str] 
            Directory of the raw, unprocessed dataset.
        """
        if val is None:
            self._raw_data_dir = self.root
            return
        if isinstance(val, str):
            val = Path(val)
        if not isinstance(val, Path):
            raise TypeError(f"raw_data_dir should be of type Path or str, but was {type(val)}")
        if self._raw_data_dir_relative_to_root:
            self._raw_data_dir = (self.root / val).resolve()
        else:
            self._raw_data_dir = val.resolve()

    def _download(
        self,
        source: str,
        destination: Union[Path, str],
        mkdir: bool = True,
        skip_if_exists: bool = False,
        extract: bool = False,
        keep_archive: bool = False,
    ):
        """Download a remote file to local filesystem

        Parameters
        ----------
        source : str
            Remote URL of the target file.
        destination : Union[Path, str]
            Full destination path and filename.
        mkdir : bool, optional
            If True, the parent directories of the destination filenames are
            made, by default True.
        skip_if_exists: bool, optional
            If True, then skip if the file already exists, by default True.
        extract: bool, optional
            If True, then extract the downloaded archive by calling
            self.extract_archive, by default False.
        keep_archive: bool, optional
            If True, and also extract is True, then the original file should be
            kept after extraction. Has no effect is extract is False.
            By default False.
        """
        if isinstance(destination, str):
            destination = Path(destination)
        if not isinstance(destination, Path):
            raise TypeError(f"Destination should be of type Path or str, but was {type(destination)}")
        if not destination.parent.exists():
            if mkdir:
                destination.parent.mkdir(parents=True)
            else:
                raise FileNotFoundError(
                    f"Parent directories of {destination} do not existand mkdir=False, so could not make them"
                )
        if skip_if_exists and destination.exists():
            return
        urllib.request.urlretrieve(source, destination)

        if extract:
            self._extract_archive(destination)

    def _batch_download(
        self,
        tasks: List[Tuple[str, Path]],
        max_attempts=3,
        retry_pause_seconds: int = 5,
        mkdir: bool = True,
        non_critical_urls: Optional[List[str]] = None,
    ):
        """Batch download remote files to local filesystem.

        Parameters
        ----------
        tasks : List[Tuple[str, Path]]
            List of tuples, with each tuple being a download task. Each task
            consists of first a remote source location string, and then
            secondarily, a destination path where the file should be saved.
            For example:
            [("http://example1.com/remote_file1","/tmp/file1.txt),
            ("http://example2.com/remote_file2","/tmp/file2.txt)]
        max_attempts : int, optional
            Maximum number of times to retry after failed download,
            by default 3.
        retry_pause_seconds : int, optional
            Pause for this length of time in seconds before download retries, by
            default 5.
        mkdir : bool, optional
            Used as an argument to the download function used by this batch
            downloader, if true, then parent directories for the named
            destination files are made if they do not exist. By default True.
        non_critical_urls: list(str), optional
            In datasets such as TCGA, some files that should exist on the server
            are absent. If after the set number of download attempts
            (max_attempts) has been reached and the file is still not
            downloaded, no exception will be raised if the remote URL is present
            in this non_critical_urls list.  By default None.
        """
        if isinstance(non_critical_urls, str):
            non_critical_urls = [non_critical_urls]
        failed_items = []
        if max_attempts == 0:
            if all(fi[0] in non_critical_urls for fi in failed_items):
                return
            else:
                raise URLError(
                    f"Cannot make any more download attempts, {len(tasks)} tasks remain, they were:\n {tasks}. Set download=False to process what was downloaded"
                )
        for source, destination in progressbar(tasks, redirect_stdout=True):
            try:
                self._download(source, destination)
            except Exception as e:
                print(e)
                print(f"Failed to get {source}, will retry ")
                failed_items.append((source, destination))
        if len(failed_items) > 0:
            sleep(retry_pause_seconds)
            self._batch_download(
                failed_items,
                max_attempts=max_attempts - 1,
                retry_pause_seconds=retry_pause_seconds,
            )

    def _processed_dataset_exists(self, required_files: List[Path]) -> bool:
        """Checks if all listed files are present in the dataset root directory

        Parameters
        ----------
        required_files : List[Path]
            List of file names to check exist in dataset root directory

        Returns
        -------
        bool
            True if all files were found in root of dataset directory
        """
        return all([(self.root / file).exists() for file in required_files])

    def _raw_dataset_exists(self, required_files: List[Path]) -> bool:
        """Checks if all listed files are present in the raw_data_dir of dataset

        Parameters
        ----------
        required_files : List[Path]
            List of file names to check exist in raw_data_dir of dataset

        Returns
        -------
        bool
            True if all files were found in raw_data_dir of dataset
        """
        return all([(self.root / file).exists() for file in required_files])

    def _extract_archive(
        self,
        archive: Union[Path, str],
        destination_dir: Optional[Union[Path, str]] = None,
        keep_original: bool = False,
    ):
        """Extract archive

        Parameters
        ----------
        archive : Union[Path, str]
            Path to archive
        destination_dir : Optional[Union[Path, str]], optional
            Directory where the archive is to be extracted to.  If None, then
            the destination is the location of the archive file,
            by default None.
        keep_original : bool
            If true, the original archive is kept.  Only comes into effect if
            the archive is a .gz, excluding .tar.gz, by default False.
        """
        if isinstance(archive, str):
            archive = Path(archive)
        if destination_dir is None:
            destination_dir = archive.parent
        if not destination_dir.exists():
            destination_dir.mkdir(parents=True)
        if not str(archive).endswith(".tar.gz"):
            with gzip.open(archive, "rb") as f_in:
                with open(str(archive)[: -len(".gz")], "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
            if not keep_original:
                Path(archive).unlink(missing_ok=True)
        else:
            shutil.unpack_archive(archive, destination_dir)

    def _call_if_file_missing(
        self,
        required_file: Union[Path, str],
        func: callable,
        func_kwargs: Optional[dict] = None,
        path_relative_to_root: bool = False,
    ):
        """If a file is missing, call the appropriate function

        Parameters
        ----------
        required_file : Union[Path, str]
            File that should be present.
        func : callable
            The function to be called if the file is absent.
        func_kwargs : dict
            kwargs to be used when calling the given function.
        path_relative_to_root : bool, optional
            If True, then the given path is relative to the root directory of
            the dataset. If false, then assume a full path has been supplied,
            by default False.
        """
        if path_relative_to_root:
            required_file = self.root / required_file
        if func_kwargs is None:
            func_kwargs = {}
        if not required_file.exists():
            func(**func_kwargs)

    def keys(self) -> list:
        """Get a list of Datasets contained within this PackagedDataset.

        Returns a list of Dataset names contained within this PackagedDataset
        and allows access to them via pds.['dataset_name'] - dictionary-like
        notation.

        Returns
        -------
        List
            List of keys which allow accessing pd.DataFrames belonging to this
            PackagedDataset.
        """
        return self.ds_keys()

    def __iter__(self):
        """Like a dictionary, iterating the object returns keys to Datasets

        Returns
        -------
        List
            List of dataset names
        """        
        return (k for k in self.ds_keys())

    def ds_keys(self) -> list:
        """Get a list of Datasets contained within this PackagedDataset.

        Returns a list of Dataset names contained within this PackagedDataset
        and allows access to them via pds.['dataset_name'] - dictionary-like
        notation.

        Returns
        -------
        List
            List of keys which allow accessing pd.DataFrames belonging to this
            PackagedDataset 
        """
        return self._keys_to_accessible_dss

    def df_keys(self) -> list:
        """Get a list of available DataFrames.

        PackagedDatasets may include DataFrames, useful in the capture of
        metadata.

        Returns
        -------
        List
            List of keys which allow accessing metadata (typically)
            pd.DataFrames belonging to this PackagedDataset.
        """
        return self._keys_to_accessible_dfs

    def register_df_key(self, key:Union[str, List[str]])->None:
        """Register a dataframe key with the PackagedDataset

        Packaged datasets may contain or have access to multiple pd.DataFrames
        which accompany the main dataset. In the case of the CMAP dataset, the
        main Phenonaut Dataset contains the L1000 values, along with features
        and metadata. The supporting dataframes contain information on the
        perturbation type, compound information, etc. Keys are typically the
        same as their HDF5 store key values, although this is up to the specific
        PackagedDataset implementation.

        Parameters
        ----------
        key : Union[str, List[str]]
            Short string which may be used to access the pd.DataFrame.

        Raises
        ------
        TypeError
            [description]
        """        
        if isinstance(key, str):
            self._keys_to_accessible_dfs.append(key)
            return
        if isinstance(key, list):
            self._keys_to_accessible_dfs.extend(key)
            return
        raise TypeError(f"Key must be a str or list, given type was {type(key)}")


    def register_ds_key(self, key:Union[str, List[str]])->None:
        """Register a Phenonaut Dataset key with the PackagedDataset

        Packaged datasets may contain or have access to multiple pd.DataFrames
        from which Phenonaut Datasets can be created. These Datasets contain not
        only the pd.DataFrame containing data, but also features and additional
        metadata such as history and origin. The main Phenonaut Dataset should
        be called, by convention "ds" and additional Phenonaut Datasets given a
        descriptive name. In the case of the CMAP dataset, the main DataSet (ds)
        contains the L1000 values, feature and metadata. Supporting dataframes
        are accessible using the .get_df("name") method and contain information
        on things like the perturbation type, compound information, etc. Keys
        are typically the same as their HDF5 store key values, although this is
        up to the specific PackagedDataset implementation.

        Parameters
        ----------
        key : Union[str, List[str]]
            Short string which may be used to access the pd.DataFrame.
            When given, to the function, the name will be registered. After
            which the inheriting class should allow access to a Phenonaut
            Dataset via the __getitem__ method, allowing access to the
            Phenonaut Dataset via pdf['ds'].

        Raises
        ------
        TypeError
            Given key must be a str or list of str.
        """        
        if isinstance(key, str):
            self._keys_to_accessible_dss.append(key)
            return
        if isinstance(key, list):
            self._keys_to_accessible_dss.extend(key)
            return
        raise TypeError(f"Key must be a str or list, given type was {type(key)}")

    def __getitem__(self, key):
        """Get dataset using ['ds'] notation

        Access Datasets using simple ['ds_name'] - dictionary-like notation. If
        a Dataset of that name is not present, then check if a DataFrame of that
        name is present.  If no Dataset or DataFrame is found, then an error is
        raised. 

        Parameters
        ----------
        key : str
            Key of dataset to be returned, if it does not exist in ds_keys, then look in dataframes

        Returns
        -------
        Union[Dataset, pd.DataFrame, pd.Series]
            Requested Dataset of pd.DataFrame

        Raises
        ------
        KeyError
            Key not found
        """        
        if key in self.ds_keys():
            return self.get_ds(key)
        if key in self.df_keys():
            return self.get_df(key)
        raise KeyError(f"The key '{key}' was not found in registered datsets or dataframes (.ds_keys(), or df_keys())")

    @abstractmethod
    def get_df(self, key:str):
        """Abstract method - Get DataFrame

        Abstract method which all inheriting classes are required to implement
        for retrieval of DataFrames.

        Parameters
        ----------
        key : str
            Name of DataFrame
        """        
        pass

    @abstractmethod
    def get_ds(self, key:str):
        """Abstract method - Get Dataset

        Abstract method which all inheriting classes are required to implement
        for retrieval of Datasets.

        Parameters
        ----------
        key : str
            Name of Dataset
        """ 
        pass

